import os
import subprocess
from pathlib import Path, PosixPath

import psycopg
import pytest

import pgai

# skip tests in this module if disabled
enable_privileges_tests = os.getenv("ENABLE_PRIVILEGES_TESTS")
if enable_privileges_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str, dbname: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/{dbname}"


def where_am_i() -> str:
    if "WHERE_AM_I" in os.environ and os.environ["WHERE_AM_I"] == "docker":
        return "docker"
    return "host"


def docker_dir() -> str:
    return str(
        PosixPath("/").joinpath(
            "pgai", "projects", "pgai", "db", "tests", "vectorizer", "privileges"
        )
    )


def host_dir() -> Path:
    return Path(__file__).parent.absolute()


def psql_file(user, dbname, file: str) -> None:
    cmd = " ".join(
        [
            "psql",
            f'''-d "{db_url(user, dbname)}"''',
            "-v ON_ERROR_STOP=1",
            "-X",
            f"-f {docker_dir()}/{file}",
        ]
    )
    if where_am_i() != "docker":
        cmd = f"docker exec -w {docker_dir()} pgai-db {cmd}"
    subprocess.run(cmd, check=True, shell=True, env=os.environ, cwd=str(host_dir()))


@pytest.fixture(scope="module", autouse=True)
def init():
    psql_file("postgres", "postgres", "init0.sql")
    pgai.install(db_url("alice", "privs"))
    psql_file("alice", "privs", "init1.sql")


def test_jill_privileges():
    psql_file("jill", "privs", "jill.sql")


def test_create_vectorizer_privileges():
    # set up role "base" and role "member", which is member of base
    with psycopg.connect(db_url("postgres", "postgres"), autocommit=True) as con:
        with con.cursor() as cur:
            cur.execute("drop database if exists vec_priv;")
            cur.execute(
                """
                drop role if exists member;
                drop role if exists base;
                create role base with login;
                create role member with login;
                grant base to member;
                """
            )
            cur.execute("create database vec_priv owner base;")
    # connect as "base", create vectorizer

    pgai.install(db_url("base", "vec_priv"))
    with psycopg.connect(db_url("base", "vec_priv")) as con:
        with con.cursor() as cur:
            cur.execute(
                """
                create table blog(id bigint primary key, content text);
                select ai.create_vectorizer(
                    'blog'
                  , loading => ai.loading_column('content')
                  , destination=>ai.destination_table('base_vectorizer')
                  , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
                  , chunking=>ai.chunking_character_text_splitter(128, 10)
                  , scheduling=>ai.scheduling_none()
                  , indexing=>ai.indexing_none()
                );
                """
            )
    # connect as "member", create vectorizer
    with psycopg.connect(db_url("member", "vec_priv")) as con:
        with con.cursor() as cur:
            cur.execute("""
                  select ai.create_vectorizer(
                    'blog'
                  , loading => ai.loading_column('content')
                  , destination=> ai.destination_table('member_vectorizer')
                  , embedding=>ai.embedding_openai('text-embedding-3-small', 768)
                  , chunking=>ai.chunking_character_text_splitter(128, 10)
                  , scheduling=>ai.scheduling_none()
                  , indexing=>ai.indexing_none()
                );
            """)



import os

import psycopg
import pytest

# skip tests in this module if disabled
enable_vectorizer_tests = os.getenv("ENABLE_VECTORIZER_TESTS")
if enable_vectorizer_tests == "0":
    pytest.skip(allow_module_level=True)


def db_url(user: str) -> str:
    return f"postgres://{user}@127.0.0.1:5432/test"


def test_embedding_openai():
    tests = [
        (
            "select ai.embedding_openai('text-embedding-3-small', 128)",
            {
                "implementation": "openai",
                "config_type": "embedding",
                "model": "text-embedding-3-small",
                "dimensions": 128,
                "api_key_name": "OPENAI_API_KEY",
            },
        ),
        (
            """select ai.embedding_openai('text-embedding-3-small', 128, chat_user=>'bob')""",
            {
                "implementation": "openai",
                "config_type": "embedding",
                "model": "text-embedding-3-small",
                "dimensions": 128,
                "user": "bob",
                "api_key_name": "OPENAI_API_KEY",
            },
        ),
        (
            "select ai.embedding_openai('text-embedding-3-small', 128, api_key_name=>'DEV_API_KEY')",
            {
                "implementation": "openai",
                "config_type": "embedding",
                "model": "text-embedding-3-small",
                "dimensions": 128,
                "api_key_name": "DEV_API_KEY",
            },
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in tests:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]


def test_embedding_voyageai():
    ok = [
        (
            "select ai.embedding_voyageai('voyage-3-lite', 512)",
            {
                "implementation": "voyageai",
                "config_type": "embedding",
                "model": "voyage-3-lite",
                "dimensions": 512,
                "input_type": "document",
                "api_key_name": "VOYAGE_API_KEY",
            },
        ),
        (
            "select ai.embedding_voyageai('voyage-3-lite', 512, api_key_name => 'TEST_API_KEY', input_type => null)",
            {
                "implementation": "voyageai",
                "config_type": "embedding",
                "model": "voyage-3-lite",
                "dimensions": 512,
                "api_key_name": "TEST_API_KEY",
            },
        ),
        (
            "select ai.embedding_voyageai('voyage-3-lite', 512, api_key_name => 'TEST_API_KEY', input_type => 'query')",
            {
                "implementation": "voyageai",
                "config_type": "embedding",
                "model": "voyage-3-lite",
                "dimensions": 512,
                "input_type": "query",
                "api_key_name": "TEST_API_KEY",
            },
        ),
    ]
    bad = [
        (
            "select ai.embedding_voyageai('voyage-3-lite', 512, input_type => 'foo')",
            'invalid input_type for voyage ai "foo"',
        ),
    ]
    with psycopg.connect(db_url("test")) as con:
        with con.cursor() as cur:
            for query, expected in ok:
                cur.execute(query)
                actual = cur.fetchone()[0]
                assert actual.keys() == expected.keys()
                for k, v in actual.items():
                    assert k in expected and v == expected[k]
            for query, err in bad:
                try:
                    cur.execute(query)
                except psycopg.ProgrammingError as ex:
                    msg = str(ex.args[0])
                    assert len(msg) >= len(err) and msg[: len(err)] == err
                else:
                    pytest.fail(f"expected exception: {err}")


def test_validate_embedding():
    ok = [
        "select ai._validate_embedding( ai.embedding_openai('text-embedding-3-small', 756))",
    ]
    bad = [
        (
            "select ai._validate_embedding(ai.chunking_character_text_splitter(128, 10))",
            "invalid config_type for embedding config",
        ),
        (
            """select ai._validate_embedding('{"config_type": "embedding"}')""",
            "embedding implementation not specified",
        ),
        (
            """select ai._validate_embedding('{"config_type": "embedding", "implementation": "bob"}')""",
            'invalid embedding implementation: "bob"',
        ),
    ]
    with psycopg.connect(db_url("test"), autocommit=True) as con:
        with con.cursor() as cur:
            for query in ok:
                cur.execute(query)
                assert True
            for query, err in bad:
                try:
                    cur.execute(query)
                except psycopg.ProgrammingError as ex:
                    msg = str(ex.args[0])
                    assert len(msg) >= len(err) and msg[: len(err)] == err
                else:
                    pytest.fail(f"expected exception: {err}")



from pathlib import Path

import psycopg
from psycopg.sql import SQL, Identifier

import pgai
from tests.semantic_catalog.utils import (
    PostgresContainer,
)


async def test_desc_injest(container: PostgresContainer):
    database = "test_desc_injest"
    container.drop_database(database)
    container.create_database(database)
    expected: set[str] = {
        "postgres_air.account",
        "postgres_air.account.account_id",
        "postgres_air.account.login",
        "postgres_air.account.first_name",
        "postgres_air.account.last_name",
        "postgres_air.account.frequent_flyer_id",
        "postgres_air.account.update_ts",
        "postgres_air.aircraft",
        "postgres_air.aircraft.model",
        "postgres_air.aircraft.range",
        "postgres_air.aircraft.class",
        "postgres_air.aircraft.velocity",
        "postgres_air.aircraft.code",
        "postgres_air.airport",
        "postgres_air.airport.airport_code",
        "postgres_air.airport.airport_name",
        "postgres_air.airport.city",
        "postgres_air.airport.airport_tz",
        "postgres_air.airport.continent",
        "postgres_air.airport.iso_country",
        "postgres_air.airport.iso_region",
        "postgres_air.airport.intnl",
        "postgres_air.airport.update_ts",
        "postgres_air.boarding_pass",
        "postgres_air.boarding_pass.pass_id",
        "postgres_air.boarding_pass.passenger_id",
        "postgres_air.boarding_pass.booking_leg_id",
        "postgres_air.boarding_pass.seat",
        "postgres_air.boarding_pass.boarding_time",
        "postgres_air.boarding_pass.precheck",
        "postgres_air.boarding_pass.update_ts",
        "postgres_air.booking",
        "postgres_air.booking.booking_id",
        "postgres_air.booking.booking_ref",
        "postgres_air.booking.booking_name",
        "postgres_air.booking.account_id",
        "postgres_air.booking.email",
        "postgres_air.booking.phone",
        "postgres_air.booking.update_ts",
        "postgres_air.booking.price",
        "postgres_air.booking_leg",
        "postgres_air.booking_leg.booking_leg_id",
        "postgres_air.booking_leg.booking_id",
        "postgres_air.booking_leg.flight_id",
        "postgres_air.booking_leg.leg_num",
        "postgres_air.booking_leg.is_returning",
        "postgres_air.booking_leg.update_ts",
        "postgres_air.events",
        "postgres_air.events.time",
        "postgres_air.events.name",
        "postgres_air.events.params",
        "postgres_air.flight",
        "postgres_air.flight.flight_id",
        "postgres_air.flight.flight_no",
        "postgres_air.flight.scheduled_departure",
        "postgres_air.flight.scheduled_arrival",
        "postgres_air.flight.departure_airport",
        "postgres_air.flight.arrival_airport",
        "postgres_air.flight.status",
        "postgres_air.flight.aircraft_code",
        "postgres_air.flight.actual_departure",
        "postgres_air.flight.actual_arrival",
        "postgres_air.flight.update_ts",
        "postgres_air.frequent_flyer",
        "postgres_air.frequent_flyer.frequent_flyer_id",
        "postgres_air.frequent_flyer.first_name",
        "postgres_air.frequent_flyer.last_name",
        "postgres_air.frequent_flyer.title",
        "postgres_air.frequent_flyer.card_num",
        "postgres_air.frequent_flyer.level",
        "postgres_air.frequent_flyer.award_points",
        "postgres_air.frequent_flyer.email",
        "postgres_air.frequent_flyer.phone",
        "postgres_air.frequent_flyer.update_ts",
        "postgres_air.hypertable_test",
        "postgres_air.hypertable_test.time",
        "postgres_air.hypertable_test.location",
        "postgres_air.hypertable_test.time_received",
        "postgres_air.hypertable_test.params",
        "postgres_air.passenger",
        "postgres_air.passenger.passenger_id",
        "postgres_air.passenger.booking_id",
        "postgres_air.passenger.booking_ref",
        "postgres_air.passenger.passenger_no",
        "postgres_air.passenger.first_name",
        "postgres_air.passenger.last_name",
        "postgres_air.passenger.account_id",
        "postgres_air.passenger.update_ts",
        "postgres_air.passenger.age",
        "postgres_air.phone",
        "postgres_air.phone.phone_id",
        "postgres_air.phone.account_id",
        "postgres_air.phone.phone",
        "postgres_air.phone.phone_type",
        "postgres_air.phone.primary_phone",
        "postgres_air.phone.update_ts",
        "postgres_air.events_daily",
        "postgres_air.events_daily.name",
        "postgres_air.events_daily.bucket",
        "postgres_air.flight_summary",
        "postgres_air.flight_summary.flight_no",
        "postgres_air.flight_summary.departure_airport",
        "postgres_air.flight_summary.arrival_airport",
        "postgres_air.flight_summary.scheduled_departure",
        "postgres_air.passenger_details",
        "postgres_air.passenger_details.first_name",
        "postgres_air.passenger_details.last_name",
        "postgres_air.passenger_details.booking_ref",
        "postgres_air.advance_air_time",
        "postgres_air.update_flight_status",
    }
    script = Path(__file__).parent.joinpath("data", "descriptions.sql").read_text()
    connection_str = container.connection_string(database=database)
    pgai.install(connection_str)
    async with await psycopg.AsyncConnection.connect(connection_str) as con:  # noqa SIM117
        async with con.cursor() as cur:
            await cur.execute("select ai.create_semantic_catalog('my_catalog')")
            row = await cur.fetchone()
            if row is None:
                raise RuntimeError("Failed to create catalog")
            catalog_id: int = int(row[0])
            await con.commit()
            for line in script.split("\n"):
                await cur.execute(line)  # pyright: ignore [reportArgumentType]
                await con.commit()
            query = SQL("""\
                select array_to_string(x.objnames, '.')
                from ai.{} x
                where x.description is not null
            """).format(Identifier(f"semantic_catalog_obj_{catalog_id}"))
            await cur.execute(query)
            actual: set[str] = set()
            async for row in cur:
                actual.add(row[0])
    assert actual == expected



"""SentenceTransformers embedding provider for the semantic catalog vectorizer.

This module implements embedding functionality using SentenceTransformers as the
embedding provider. It provides functions for embedding both batches of content and
individual queries.
"""

import logging
from collections.abc import Sequence
from contextlib import contextmanager

import numpy as np
import numpy.typing as npt
from sentence_transformers import SentenceTransformer

from pgai.semantic_catalog.vectorizer import SentenceTransformersConfig
from pgai.semantic_catalog.vectorizer.models import EmbedRow


@contextmanager
def disable_logging():
    """
    Disable logging for the sentence_transformers and transformers_modules
    libraries.
    """
    st_logger = logging.getLogger("sentence_transformers")
    st_prev_level = st_logger.level
    st_logger.setLevel(logging.ERROR)
    t_logger = logging.getLogger("transformers_modules")
    t_prev_level = t_logger.level
    t_logger.setLevel(logging.ERROR)

    try:
        yield
    finally:
        st_logger.setLevel(st_prev_level)
        t_logger.setLevel(t_prev_level)


async def embed_batch(
    config: SentenceTransformersConfig, batch: list[EmbedRow]
) -> None:
    """Generate embeddings for a batch of content using SentenceTransformers.

    Creates vector embeddings for multiple items using SentenceTransformers and
    updates the vector field in each EmbedRow object with the resulting embedding.

    Args:
        config: Configuration for the SentenceTransformers embedding service.
        batch: List of EmbedRow objects containing content to be embedded.

    Raises:
        AssertionError: If the number of embeddings returned doesn't match the batch size.
    """
    st = SentenceTransformer(
        config.model, trust_remote_code=True
    )  # TODO: configurable?
    sentences: list[str] = [x.content for x in batch]
    with disable_logging():
        results: npt.NDArray[np.float64] = st.encode(  # pyright: ignore [reportUnknownMemberType,reportUnknownVariableType]
            sentences=sentences,
            prompt_name=None,
            prompt=None,
            batch_size=len(batch),
            show_progress_bar=False,
            output_value="sentence_embedding",
            precision="float32",
            convert_to_numpy=True,
            convert_to_tensor=False,
            device=None,
            normalize_embeddings=True,
        )
    assert len(results) == len(batch)
    for i, row in enumerate(results):
        batch[i].vector = list(row)


async def embed_query(
    config: SentenceTransformersConfig, query: str
) -> Sequence[float]:
    """Generate an embedding for a single query using SentenceTransformers.

    Creates a vector embedding for a query string using SentenceTransformers.

    Args:
        config: Configuration for the SentenceTransformers embedding service.
        query: The query string to embed.

    Returns:
        A vector embedding (sequence of floats) for the query.

    Raises:
        AssertionError: If the number of embeddings returned is not exactly 1.
    """
    st = SentenceTransformer(
        config.model, trust_remote_code=True
    )  # TODO: configurable?
    sentences: list[str] = [query]
    with disable_logging():
        results: npt.NDArray[np.float64] = st.encode(  # pyright: ignore [reportUnknownMemberType,reportUnknownVariableType]
            sentences=sentences,
            prompt_name=None,
            prompt=None,
            batch_size=1,
            show_progress_bar=False,
            output_value="sentence_embedding",
            precision="float32",
            convert_to_numpy=True,
            convert_to_tensor=False,
            device=None,
            normalize_embeddings=True,
        )
    assert len(results) == 1
    return list(results[0])



import os
from collections.abc import Generator
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

import pytest
from psycopg import Connection
from psycopg.rows import dict_row
from testcontainers.localstack import LocalStackContainer  # type: ignore

from pgai.vectorizer.parsing import DOCLING_CACHE_DIR
from tests.vectorizer.cli.conftest import (
    TestDatabase,
    configure_vectorizer,
    download_docling_models,
    run_vectorizer_worker,
)

docs = [
    "Sacred Texts of PostgreSQL.pdf",
    "sample_pdf.pdf",
    "sample_with_table.pdf",
    "sample_txt.txt",
    "basic-v3plus2.epub",
    "test.md",
    "lego_sets.pdf",
    "text_screenshot.png",
    "electromagnetic_radiation.docx",
]


@pytest.fixture(scope="session")
def s3_bucket() -> Generator[str, None, None]:
    region = "eu-central-1"
    localstack = LocalStackContainer(
        image="localstack/localstack:4", region_name=region
    ).with_services("s3")  # type: ignore
    # hardcoding port so vcr cassettes are always matching the uri
    localstack.ports[localstack.edge_port] = 32882  # type: ignore
    localstack.start()

    # Set environment variables for any lib that uses boto3 (such as smart-open)
    os.environ["AWS_ENDPOINT_URL"] = localstack.get_url()
    os.environ["AWS_SHARED_CREDENTIALS_FILE"] = "/dev/null"
    os.environ["AWS_ACCESS_KEY_ID"] = "testcontainers-localstack"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testcontainers-localstack"

    s3_test_bucket_name = "localstack-test-bucket"
    s3 = localstack.get_client("s3")  # type: ignore

    s3.create_bucket(  # type: ignore
        Bucket=s3_test_bucket_name,
        CreateBucketConfiguration={"LocationConstraint": region},
    )
    for doc in docs:
        s3.upload_file(  # type: ignore
            Path(__file__).parent / "documents" / doc,
            s3_test_bucket_name,
            doc,
        )

    yield s3_test_bucket_name
    localstack.stop()


def setup_documents_table(
    connection: Connection,
    number_of_rows: int,
    base_path: Path | str,
    documents: list[str],
):
    table_name = "documents"
    with connection.cursor(row_factory=dict_row) as cur:
        # Create source table
        cur.execute(f"""
                CREATE TABLE {table_name} (
                    id INT NOT NULL PRIMARY KEY,
                    url TEXT NOT NULL
                )
            """)

        values: list[tuple[int, str]] = []
        for i in range(1, number_of_rows + 1):
            doc_index = (i - 1) % len(documents)
            s3_url = str(base_path) + "/" + documents[doc_index]
            values.append((i, str(s3_url)))

        cur.executemany("INSERT INTO documents (id, url) VALUES (%s, %s)", values)
    return table_name


def configure_document_vectorizer(
    connection: Connection,
    openai_proxy_url: str | None = None,
    number_of_rows: int = 1,
    concurrency: int = 1,
    batch_size: int = 1,
    base_path: Path | str = Path(__file__).parent / "documents",
    loading: str = "ai.loading_uri(column_name => 'url')",
    parsing: str = "ai.parsing_auto()",
    chunking: str = "chunking_recursive_character_text_splitter(chunk_size => 700,"
    # ' | ' is a separator for the md table extracted from lego_sets.pdf
    # otherwise we reach max token length
    "separators => array[E'\n\n', E'\n', '.', '?', '!', ' ', '', ' | '])",
    formatting: str = "formatting_python_template('$chunk')",
    documents: list[str] | None = None,
) -> int:
    """Creates and configures a vectorizer for testing"""
    if documents is None:
        documents = docs
    documents_table = setup_documents_table(
        connection, number_of_rows, base_path, documents
    )
    base_url = (
        f", base_url => '{openai_proxy_url}'" if openai_proxy_url is not None else ""
    )
    embedding = f"embedding_openai('text-embedding-ada-002', 1536{base_url})"
    return configure_vectorizer(
        documents_table,
        connection,
        concurrency=concurrency,
        batch_size=batch_size,
        chunking=chunking,
        formatting=formatting,
        embedding=embedding,
        loading=loading,
        parsing=parsing,
    )


def list_dir_contents(path: Path):
    """Return a sorted list of relative paths (files and dirs) in the directory."""
    return sorted(p.relative_to(path) for p in path.rglob("*") if p.exists())


def test_simple_document_embedding_local(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    # First download the models used by Docling to the cache dir
    download_docling_models()

    # grab a snapshot of the docling and EasyOCR models dir to compare later
    easyocr_models_dir = Path.home().joinpath(".EasyOCR/model")
    docling_models_dir_before = list_dir_contents(DOCLING_CACHE_DIR)
    easyocr_models_dir_before = list_dir_contents(easyocr_models_dir)

    """Test that a document is successfully embedded"""
    connection = cli_db[1]
    vectorizer_id = configure_document_vectorizer(cli_db[1], number_of_rows=len(docs))

    # Huggingface.co hosts is being ignored globally.
    # Docling will download its models if not found.
    with vcr_.use_cassette("simple-docs.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    if result.exception:
        print(result.stdout)

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) as errors FROM ai.vectorizer_errors;")
        assert cur.fetchone()["errors"] == 0  # type: ignore

        cur.execute("SELECT count(*) as count FROM documents_embedding_store;")
        assert cur.fetchone()["count"] > len(docs)  # type: ignore

        cur.execute("SELECT chunk FROM documents_embedding_store;")
        chunks = cur.fetchall()
        chunks_str = "\n".join([chunk["chunk"] for chunk in chunks])

        # Sacred Texts of PostgreSQL.pdf
        assert "And lo, there came forth PostgreSQL, blessed be its name" in chunks_str

        # sample_pdf.pdf
        assert "Maecenas mauris lectus" in chunks_str

        # sample_with_table.pdf
        assert "This is an example of a data table." in chunks_str

        # sample_txt.txt
        assert "Fromage frais cheese and biscuits danish fontina" in chunks_str

        # basic-v3plus2.epub
        # epub is not supported by docling but by pymupdf. auto parser handles this.
        assert "Wants pawn term dare worsted ladle gull hoe lift" in chunks_str

        # test.md
        assert "Hello I am a test md document" in chunks_str

        # lego_sets.pdf
        assert "7190-1 Millennium Falcon" in chunks_str

        # text_screenshot.png
        assert "Use cases include providing chatbot" in chunks_str

        # electromagnetic_radiation.docx
        assert "All forms of EMR travel at the speed of light in a vacuum" in chunks_str

    # ensure no extra models were downloaded during the test
    assert list_dir_contents(DOCLING_CACHE_DIR) == docling_models_dir_before
    assert list_dir_contents(easyocr_models_dir) == easyocr_models_dir_before


def test_simple_document_embedding_s3_no_credentials(
    s3_bucket: str,
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
):
    """Test that embedding fails when no credentials are available"""
    vectorizer_id = configure_document_vectorizer(
        cli_db[1], base_path=f"s3://{s3_bucket}"
    )

    if "AWS_ACCESS_KEY_ID" in os.environ:
        del os.environ["AWS_ACCESS_KEY_ID"]
    if "AWS_SECRET_ACCESS_KEY" in os.environ:
        del os.environ["AWS_SECRET_ACCESS_KEY"]

    # This is necessary to prevent boto3 from using the default credentials file.
    os.environ["AWS_SHARED_CREDENTIALS_FILE"] = "/dev/null"

    # No cassette because it should never get to an API call.
    run_vectorizer_worker(cli_db_url, vectorizer_id)

    connection = cli_db[1]
    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) FROM ai._vectorizer_q_1;")
        assert cur.fetchone()["count"] == 1  # type: ignore
        cur.execute("SELECT id, message, details FROM ai.vectorizer_errors")
        records = cur.fetchall()
        assert len(records) == 1
        error = records[0]
        assert error["id"] == vectorizer_id
        assert error["message"] == "loading failed"
        assert error["details"] == {
            "loader": "uri",
            "error_reason": "Unable to locate credentials",
            "is_retryable": True,
        }


def test_simple_document_embedding_s3(
    s3_bucket: str,
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """Test that a document is successfully embedded"""
    connection = cli_db[1]
    vectorizer_id = configure_document_vectorizer(
        cli_db[1], base_path=f"s3://{s3_bucket}"
    )

    # required for accessing this S3 bucket in particular
    # os.environ["AWS_DEFAULT_REGION"] = "eu-central-1"

    # This is necessary to prevent boto3 from pulling credentials
    # from WS Instance Metadata Service when those are missing.
    # Comment if you need to recreate the cassette.
    # os.environ["AWS_ACCESS_KEY_ID"] = "FAKE"
    # os.environ["AWS_SECRET_ACCESS_KEY"] = "FAKE"
    os.environ["AWS_SHARED_CREDENTIALS_FILE"] = "/dev/null"

    with vcr_.use_cassette("simple-s3-docs.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    if result.exception:
        print(result.stdout)

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) as count FROM documents_embedding_store;")
        assert cur.fetchone()["count"] > 0  # type: ignore
        cur.execute("SELECT chunk FROM documents_embedding_store;")
        chunks = cur.fetchall()
        chunks_str = "\n".join([chunk["chunk"] for chunk in chunks])
        assert "And lo, there came forth PostgreSQL, blessed be its name" in chunks_str


def test_binary_document_embedding(
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """Test that a document stored as binary data is successfully embedded"""
    connection = cli_db[1]

    # Create a table with bytea column
    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("""
            CREATE TABLE binary_documents (
                id INT NOT NULL PRIMARY KEY,
                byte_content BYTEA NOT NULL
            )
        """)

        # Read PDF files and insert as binary
        base_path = Path(__file__).parent / "documents"
        values: list[tuple[int, bytes]] = []
        for i, doc in enumerate(docs, start=1):
            if doc.endswith(".pdf"):  # Only process PDF files
                file_path = base_path / doc
                with open(file_path, "rb") as f:
                    binary_content = f.read()
                values.append((i, binary_content))

        cur.executemany(
            "INSERT INTO binary_documents (id, byte_content) VALUES (%s, %s)", values
        )

    # Configure vectorizer with binary loading
    vectorizer_id = configure_vectorizer(
        "binary_documents",
        connection,
        loading="ai.loading_column(column_name => 'byte_content')",
        parsing="ai.parsing_auto()",
        chunking="chunking_character_text_splitter()",
        formatting="formatting_python_template('$chunk')",
    )

    with vcr_.use_cassette("binary-docs.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    if result.exception:
        print(result.stdout)

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) as count FROM binary_documents_embedding_store;")
        count = cur.fetchone()["count"]  # type: ignore
        assert count > 0  # We should have some chunks

        cur.execute("SELECT chunk FROM binary_documents_embedding_store;")
        chunks = cur.fetchall()
        chunks_str = "\n".join([chunk["chunk"] for chunk in chunks])

        # Verify content from PDFs was processed
        assert (
            "And lo, there came forth PostgreSQL, blessed be its name" in chunks_str
        )  # From Sacred Texts
        assert "Maecenas mauris lectus" in chunks_str  # From sample_pdf
        assert (
            "This is an example of a data table." in chunks_str
        )  # From sample_with_table


def test_retries_on_not_present_document_embedding_s3(
    s3_bucket: str,
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """Test that embedding documents are successfully retried"""
    connection = cli_db[1]
    vectorizer_id = configure_document_vectorizer(
        cli_db[1], base_path=f"s3://{s3_bucket}", documents=["non_existing_doc.pdf"]
    )

    with vcr_.use_cassette("doc_retries_s3_not_found.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    if result.exception:
        print(result.stdout)

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) FROM ai._vectorizer_q_1;")
        assert cur.fetchone()["count"] == 1  # type: ignore
        cur.execute("SELECT id, message, details FROM ai.vectorizer_errors")
        records = cur.fetchall()
        assert len(records) == 1
        error = records[0]
        assert error["id"] == vectorizer_id
        assert error["message"] == "loading failed"
        assert error["details"] == {
            "loader": "uri",
            "error_reason": f"unable to access bucket: '{s3_bucket}'"
            " key: 'non_existing_doc.pdf' version: None"
            " error: An error occurred (NoSuchKey) when"
            " calling the GetObject operation: The specified"
            " key does not exist.",
            "is_retryable": True,
        }

        cur.execute(
            "SELECT loading_retries, loading_retry_after" " FROM ai._vectorizer_q_1;"
        )
        queue_item = cur.fetchone()
        assert queue_item["loading_retries"] == 1  # type: ignore
        assert queue_item["loading_retry_after"] > datetime.now(tz=timezone.utc)  # type: ignore


def test_there_will_be_no_more_retries_after_the_sixth_failure(
    s3_bucket: str,
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """Test that embedding documents are successfully retried"""
    connection = cli_db[1]
    vectorizer_id = configure_document_vectorizer(
        cli_db[1], base_path=f"s3://{s3_bucket}", documents=["non_existing_doc.pdf"]
    )

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute(
            "UPDATE ai._vectorizer_q_1"
            " SET loading_retries=6, loading_retry_after=now() - interval '1 minute';"
        )

    with vcr_.use_cassette("doc_retries_s3_not_found.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    if result.exception:
        print(result.stdout)

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT id, message, details FROM ai.vectorizer_errors")
        records = cur.fetchall()
        assert len(records) == 1
        error = records[0]
        assert error["id"] == vectorizer_id
        assert error["message"] == "loading failed"
        assert error["details"] == {
            "loader": "uri",
            "error_reason": f"unable to access bucket: '{s3_bucket}'"
            " key: 'non_existing_doc.pdf' version: None"
            " error: An error occurred (NoSuchKey) when"
            " calling the GetObject operation: The specified"
            " key does not exist.",
            "is_retryable": False,
        }

        cur.execute(
            "SELECT loading_retries, loading_retry_after" " FROM ai._vectorizer_q_1;"
        )
        assert cur.fetchone() is None  # type: ignore

        cur.execute("SELECT count(*) FROM ai._vectorizer_q_failed_1;")
        assert cur.fetchone()["count"] == 1  # type: ignore


def test_retries_should_do_nothing_if_retry_after_is_in_the_future(
    s3_bucket: str,
    cli_db: tuple[TestDatabase, Connection],
    cli_db_url: str,
    vcr_: Any,
):
    """Test that embedding documents are successfully retried"""
    connection = cli_db[1]
    vectorizer_id = configure_document_vectorizer(
        cli_db[1], base_path=f"s3://{s3_bucket}", documents=["non_existing_doc.pdf"]
    )

    with vcr_.use_cassette("doc_retries_s3_not_found.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    if result.exception:
        print(result.stdout)

    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute("SELECT count(*) FROM ai._vectorizer_q_1;")
        assert cur.fetchone()["count"] == 1  # type: ignore
        cur.execute("SELECT id, message, details FROM ai.vectorizer_errors")
        records = cur.fetchall()
        assert len(records) == 1
        error = records[0]
        assert error["id"] == vectorizer_id
        assert error["message"] == "loading failed"
        assert error["details"] == {
            "loader": "uri",
            "error_reason": f"unable to access bucket: '{s3_bucket}'"
            " key: 'non_existing_doc.pdf' version: None"
            " error: An error occurred (NoSuchKey) when"
            " calling the GetObject operation: The specified"
            " key does not exist.",
            "is_retryable": True,
        }

        cur.execute(
            "UPDATE ai._vectorizer_q_1 "
            "SET loading_retry_after=now() + interval '10 minute';"
        )
        cur.execute(
            "SELECT loading_retries, loading_retry_after " "FROM ai._vectorizer_q_1;"
        )
        queue_item = cur.fetchone()
        assert queue_item["loading_retries"] == 1  # type: ignore
        assert queue_item["loading_retry_after"] > datetime.now(tz=timezone.utc)  # type: ignore

    with vcr_.use_cassette("doc_retries_s3_not_found.yaml"):
        result = run_vectorizer_worker(cli_db_url, vectorizer_id)

    if result.exception:
        print(result.stdout)

    # loading_retries shouldn't have changed, given that
    # the loading_retry_after field is in the future.
    with connection.cursor(row_factory=dict_row) as cur:
        cur.execute(
            "SELECT loading_retries, loading_retry_after " "FROM ai._vectorizer_q_1;"
        )
        queue_item = cur.fetchone()
        assert queue_item["loading_retries"] == 1  # type: ignore
        assert queue_item["loading_retry_after"] > datetime.now(tz=timezone.utc)  # type: ignore



from pathlib import Path

from pgai.semantic_catalog import render

from .utils import get_procedures, get_tables, get_views


def test_render_tables():
    tables = get_tables()
    for table in tables:
        table.id = 0
    actual = render.render_tables(tables)
    Path(__file__).parent.joinpath("data", "render_tables.actual").write_text(actual)
    expected = (
        Path(__file__).parent.joinpath("data", "render_tables.expected").read_text()
    )
    assert actual == expected


def test_render_views():
    views = get_views()
    for view in views:
        view.id = 0
    actual = render.render_views(views)
    Path(__file__).parent.joinpath("data", "render_views.actual").write_text(actual)
    expected = (
        Path(__file__).parent.joinpath("data", "render_views.expected").read_text()
    )
    assert actual == expected


def test_render_procedures():
    procedures = get_procedures()
    for procedure in procedures:
        procedure.id = 0
    actual = render.render_procedures(procedures)
    Path(__file__).parent.joinpath("data", "render_procedures.actual").write_text(
        actual
    )
    expected = (
        Path(__file__).parent.joinpath("data", "render_procedures.expected").read_text()
    )
    assert actual == expected



"""OpenAI embedding provider for the semantic catalog vectorizer.

This module implements embedding functionality using OpenAI as the embedding provider.
It provides functions for embedding both batches of content and individual queries.
"""

import os
from collections.abc import Sequence

from openai import AsyncClient

from pgai.semantic_catalog.vectorizer import OpenAIConfig
from pgai.semantic_catalog.vectorizer.models import EmbedRow


async def embed_batch(config: OpenAIConfig, batch: list[EmbedRow]) -> None:
    """Generate embeddings for a batch of content using OpenAI.

    Creates vector embeddings for multiple items using the OpenAI API and
    updates the vector field in each EmbedRow object with the resulting embedding.

    Args:
        config: Configuration for the OpenAI embedding service.
        batch: List of EmbedRow objects containing content to be embedded.

    Raises:
        RuntimeError: If the number of embeddings returned doesn't match the batch size.
    """
    api_key: str | None = None
    if config.api_key_name is not None:
        api_key = os.getenv(config.api_key_name)
    client = AsyncClient(api_key=api_key, base_url=config.base_url)  # TODO: cache this?
    response = await client.embeddings.create(
        input=[x.content for x in batch],
        model=config.model,
        dimensions=config.dimensions,
    )
    if len(response.data) != len(batch):
        raise RuntimeError(
            f"{len(batch)} items sent to openai but {len(response.data)} embeddings returned"
        )
    for emb in response.data:
        batch[emb.index].vector = emb.embedding


async def embed_query(config: OpenAIConfig, query: str) -> Sequence[float]:
    """Generate an embedding for a single query using OpenAI.

    Creates a vector embedding for a query string using the OpenAI API.

    Args:
        config: Configuration for the OpenAI embedding service.
        query: The query string to embed.

    Returns:
        A vector embedding (sequence of floats) for the query.

    Raises:
        RuntimeError: If the number of embeddings returned is not exactly 1.
    """
    api_key: str | None = None
    if config.api_key_name is not None:
        api_key = os.getenv(config.api_key_name)
    client = AsyncClient(api_key=api_key, base_url=config.base_url)  # TODO: cache this?
    response = await client.embeddings.create(
        input=query,
        model=config.model,
        dimensions=config.dimensions,
    )
    if len(response.data) != 1:
        raise RuntimeError(f"expected 1 embedding but got {len(response.data)}")
    return response.data[0].embedding



"""Module for generating SQL statements using AI and semantic catalog context.

This module provides functionality to generate valid SQL statements based on natural
language prompts and database context from the semantic catalog. It uses semantic search
to find relevant database objects, SQL examples, and facts, and employs an AI model to
generate SQL that fulfills the user's request. The generated SQL is validated against
the target database to ensure correctness.
"""

import logging
from dataclasses import dataclass
from typing import Any, Literal, cast

import psycopg
from jinja2 import Template
from psycopg.rows import dict_row
from psycopg.sql import SQL, Identifier
from pydantic_ai.direct import model_request
from pydantic_ai.messages import (
    ModelRequest,
    ModelResponse,
    SystemPromptPart,
    TextPart,
    ToolCallPart,
    UserPromptPart,
)
from pydantic_ai.models import KnownModelName, Model, ModelRequestParameters
from pydantic_ai.settings import ModelSettings
from pydantic_ai.tools import ToolDefinition
from pydantic_ai.usage import Usage, UsageLimits

from pgai.semantic_catalog import loader, render, search, templates
from pgai.semantic_catalog.models import (
    Fact,
    ObjectDescription,
    Procedure,
    SQLExample,
    Table,
    View,
)
from pgai.semantic_catalog.vectorizer import EmbeddingConfig, vectorizer

logger = logging.getLogger(__name__)

_template_system_prompt: Template = templates.env.get_template(
    "prompt_gen_sql_system.j2"
)
_template_user_prompt: Template = templates.env.get_template("prompt_gen_sql_user.j2")


async def _get_database_version(target_con: psycopg.AsyncConnection) -> int | None:
    """Get the major version number of the PostgreSQL database.

    Queries the database to retrieve its version number, which is used to ensure
    that generated SQL is compatible with the specific PostgreSQL version in use.

    Args:
        target_con: Asynchronous connection to the target database.

    Returns:
        The major version number (e.g., 15 for PostgreSQL 15.x) or None if unavailable.
    """
    async with target_con.cursor() as cur:
        await cur.execute("select current_setting('server_version_num', true)")
        row = await cur.fetchone()
        return int(row[0]) // 10000 if row else None


@dataclass
class DatabaseContext:
    """Container for database objects, SQL examples, facts, and their rendered representations.

    This class stores references to database objects (tables, views, procedures),
    SQL examples, and facts that are relevant to a user query, along with their
    rendered text representations for use in prompt construction.

    Attributes:
        objects: Dictionary mapping object IDs to database objects (Tables, Views, Procedures).
        sql_examples: Dictionary mapping SQL example IDs to SQLExample objects.
        facts: Dictionary mapping fact IDs to Fact objects.
        rendered_objects: Dictionary mapping object IDs to their rendered text representations.
        rendered_sql_examples: Dictionary mapping SQL example IDs to their rendered text representations.
        rendered_facts: Dictionary mapping fact IDs to their rendered text representations.
    """

    objects: dict[int, Table | View | Procedure]
    sql_examples: dict[int, SQLExample]
    facts: dict[int, Fact]
    rendered_objects: dict[int, str]
    rendered_sql_examples: dict[int, str]
    rendered_facts: dict[int, str]


async def fetch_database_context(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    catalog_id: int,
    embedding_name: str,
    embedding_config: EmbeddingConfig,
    prompt: str,
    ctx: DatabaseContext | None = None,
    sample_size: int = 3,
) -> DatabaseContext:
    """Fetch database context relevant to a prompt using semantic search.

    Performs semantic search in the catalog to find database objects, SQL examples,
    and facts that are relevant to the given prompt. The retrieved items are added to
    the provided context (or a new context is created if none is provided).

    Args:
        catalog_con: Connection to the semantic catalog database.
        target_con: Connection to the target database (where the objects are defined).
        catalog_id: ID of the semantic catalog to search in.
        embedding_name: Name of the embedding column to use for semantic search.
        embedding_config: Configuration for the embedding model.
        prompt: The natural language prompt to search for relevant context.
        ctx: Optional existing DatabaseContext to add to (None creates a new one).
        sample_size: Number of sample rows to include for tables and views (default: 3).

    Returns:
        A DatabaseContext containing the relevant database objects, SQL examples, and facts.
    """
    ctx = (
        ctx
        if ctx
        else DatabaseContext(
            objects={},
            sql_examples={},
            facts={},
            rendered_objects={},
            rendered_sql_examples={},
            rendered_facts={},
        )
    )

    emb_prompt = await vectorizer.vectorize_query(embedding_config, prompt)

    # objects
    object_descs = {
        x.id: x
        for x in await search.search_objects(
            catalog_con,
            catalog_id,
            embedding_name,
            embedding_config,
            emb_prompt,
            exclude_ids=[k for k in ctx.objects],
        )
    }
    missing_object_ids = object_descs.keys() - ctx.objects.keys()
    if missing_object_ids:
        objects = await loader.load_objects(
            catalog_con,
            target_con,
            catalog_id,
            [object_descs[id] for id in missing_object_ids],
            sample_size,
        )
        ctx.objects.update({x.id: x for x in objects})
        ctx.rendered_objects.update({x.id: render.render_object(x) for x in objects})

    # sql examples
    for x in await search.search_sql_examples(
        catalog_con,
        catalog_id,
        embedding_name,
        embedding_config,
        emb_prompt,
        exclude_ids=[k for k in ctx.sql_examples],
    ):
        if x.id in ctx.sql_examples:
            continue
        ctx.sql_examples[x.id] = x
        ctx.rendered_sql_examples[x.id] = render.render_sql_example(x)

    # facts
    for x in await search.search_facts(
        catalog_con,
        catalog_id,
        embedding_name,
        embedding_config,
        emb_prompt,
        exclude_ids=[k for k in ctx.rendered_facts],
    ):
        if x.id in ctx.facts:
            continue
        ctx.facts[x.id] = x
        ctx.rendered_facts[x.id] = render.render_fact(x)

    return ctx


async def fetch_database_context_alt(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    catalog_id: int,
    obj_ids: list[int] | None = None,
    sql_ids: list[int] | None = None,
    fact_ids: list[int] | None = None,
    sample_size: int = 3,
) -> DatabaseContext:
    """Fetch database context by explicit IDs or fetch all items from the catalog.

    Retrieves database objects, SQL examples, and facts from the semantic catalog based
    on either explicit IDs or by fetching all items if no IDs are provided.

    Args:
        catalog_con: Connection to the semantic catalog database.
        target_con: Connection to the target database (where the objects are defined).
        catalog_id: ID of the semantic catalog to fetch from.
        obj_ids: Optional list of object IDs to fetch (None fetches all).
        sql_ids: Optional list of SQL example IDs to fetch (None fetches all).
        fact_ids: Optional list of fact IDs to fetch (None fetches all).
        sample_size: Number of sample rows to include for tables and views (default: 3).

    Returns:
        A DatabaseContext containing the specified database objects, SQL examples, and facts.
    """
    ctx = DatabaseContext(
        objects={},
        sql_examples={},
        facts={},
        rendered_objects={},
        rendered_sql_examples={},
        rendered_facts={},
    )
    async with catalog_con.cursor(row_factory=dict_row) as cur:
        # objects
        sql = SQL("""\
            select x.*
            from ai.{} x
            {}
        """).format(
            Identifier(f"semantic_catalog_obj_{catalog_id}"),
            SQL("where x.id = any(%s)") if obj_ids is not None else SQL(""),
        )
        await cur.execute(sql, (obj_ids,) if obj_ids is not None else ())
        objects: list[ObjectDescription] = []
        for row in await cur.fetchall():
            objects.append(ObjectDescription(**row))
        ctx.objects = {
            x.id: x
            for x in await loader.load_objects(
                catalog_con, target_con, catalog_id, objects, sample_size
            )
        }
        ctx.rendered_objects = {
            x.id: render.render_object(x) for x in ctx.objects.values()
        }

        # sql examples
        sql = SQL("""\
            select x.*
            from ai.{} x
            {}
        """).format(
            Identifier(f"semantic_catalog_sql_{catalog_id}"),
            SQL("where x.id = any(%s)") if sql_ids is not None else SQL(""),
        )
        await cur.execute(sql, (sql_ids,) if sql_ids is not None else ())
        sql_examples: list[SQLExample] = []
        for row in await cur.fetchall():
            sql_examples.append(SQLExample(**row))
        ctx.sql_examples = {x.id: x for x in sql_examples}
        ctx.rendered_sql_examples = {
            x.id: render.render_sql_example(x) for x in sql_examples
        }

        # facts
        sql = SQL("""\
            select x.*
            from ai.{} x
            {}
        """).format(
            Identifier(f"semantic_catalog_fact_{catalog_id}"),
            SQL("where x.id = any(%s)") if fact_ids is not None else SQL(""),
        )
        await cur.execute(sql, (fact_ids,) if fact_ids is not None else ())
        facts: list[Fact] = []
        for row in await cur.fetchall():
            facts.append(Fact(**row))
        ctx.facts = {x.id: x for x in facts}
        ctx.rendered_facts = {x.id: render.render_fact(x) for x in facts}

    return ctx


def diagnostic_to_str(d: psycopg.errors.Diagnostic) -> str | None:
    msgs: list[str] = []
    if d.message_primary:
        msgs.append(d.message_primary)
    if d.message_detail:
        msgs.append(f"DETAIL: {d.message_detail}")
    if d.message_hint:
        msgs.append(f"HINT: {d.message_hint}")
    if d.statement_position:
        msgs.append(f"STATEMENT POSITION: {d.statement_position}")
    if d.context:
        msgs.append(f"CONTEXT: {d.context}")
    if d.schema_name:
        msgs.append(f"SCHEMA NAME: {d.schema_name}")
    if d.table_name:
        msgs.append(f"TABLE NAME: {d.table_name}")
    if d.column_name:
        msgs.append(f"COLUMN NAME: {d.column_name}")
    if d.constraint_name:
        msgs.append(f"CONSTRAINT NAME: {d.constraint_name}")
    if d.sqlstate:
        msgs.append(f"SQLSTATE: {d.sqlstate}")
    msg = "\n".join(msgs).strip()
    return msg if msg != "" else None


async def validate_sql_statement(
    target_con: psycopg.AsyncConnection,
    sql_statement: str,
) -> tuple[dict[str, Any] | None, str | None]:
    """Validate a SQL statement against the database.

    Attempts to execute an EXPLAIN command for the SQL statement to verify that it is
    syntactically correct and can be executed on the target database. This is done in
    a transaction that is rolled back to prevent any modifications to the database.

    Args:
        target_con: Connection to the target database.
        sql_statement: The SQL statement to validate.

    Returns:
        A tuple containing:
        - The query plan as a dictionary if validation succeeds, None otherwise.
        - The error message if validation fails, None otherwise.
    """
    async with (
        target_con.cursor(row_factory=dict_row) as cur,
        target_con.transaction(force_rollback=True) as _,
    ):
        try:
            await cur.execute(f"explain (format json) {sql_statement}")  # pyright: ignore [reportArgumentType]
            plan = await cur.fetchone()
            assert plan is not None, "explain did not return a query plan"
            logger.info("sql statement is valid")
            return plan, None
        except psycopg.Error as e:
            msg = diagnostic_to_str(e.diag) or str(e)
            logger.info(f"sql statement is invalid {msg}")
            return None, msg


@dataclass
class GenerateSQLResponse:
    """Response object for the generate_sql function.

    Contains the generated SQL statement, the context used to generate it,
    the query plan, and additional information about the generation process.

    Attributes:
        sql_statement: The generated SQL statement.
        context: The database context used to generate the SQL statement.
        command_type: The type of SQL statement generated (e.g. SELECT, INSERT, UPDATE)
        query_plan: The PostgreSQL query plan for the generated SQL statement.
        messages: List of all messages exchanged during the generation process, where the
            ModelRequest has two parts: (SystemPromptPart, UserPromptPart).
        usage: Usage statistics for the AI model calls.
    """

    sql_statement: str
    context: DatabaseContext
    command_type: str
    query_plan: dict[str, Any]
    messages: list[tuple[ModelRequest, ModelResponse]]
    usage: Usage


ContextMode = Literal["semantic_search", "entire_catalog", "specific_ids"]


async def initialize_database_context(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    catalog_id: int,
    embedding_name: str,
    embedding_config: EmbeddingConfig,
    prompt: str,
    sample_size: int = 3,
    context_mode: ContextMode = "semantic_search",
    obj_ids: list[int] | None = None,
    sql_ids: list[int] | None = None,
    fact_ids: list[int] | None = None,
) -> DatabaseContext:
    """Initialize database context based on the specified context mode.

    This function serves as a dispatcher to initialize the database context using
    one of three strategies:
    1. Semantic search: Find relevant items based on the prompt.
    2. Entire catalog: Fetch all items from the catalog.
    3. Specific IDs: Fetch items with the specified IDs.

    Args:
        catalog_con: Connection to the semantic catalog database.
        target_con: Connection to the target database.
        catalog_id: ID of the semantic catalog.
        embedding_name: Name of the embedding column to use for semantic search.
        embedding_config: Configuration for the embedding model.
        prompt: The natural language prompt to search for relevant context.
        sample_size: Number of sample rows to include for tables and views (default: 3).
        context_mode: The mode to use for initializing the context (default: "semantic_search").
        obj_ids: Optional list of object IDs to fetch (for "specific_ids" mode).
        sql_ids: Optional list of SQL example IDs to fetch (for "specific_ids" mode).
        fact_ids: Optional list of fact IDs to fetch (for "specific_ids" mode).

    Returns:
        A DatabaseContext based on the specified context mode.

    Raises:
        AssertionError: If the context_mode is not one of the supported values.
    """
    assert context_mode in {"semantic_search", "entire_catalog", "specific_ids"}
    match context_mode:
        case "semantic_search":
            return await fetch_database_context(
                catalog_con,
                target_con,
                catalog_id,
                embedding_name,
                embedding_config,
                prompt,
                ctx=None,
                sample_size=sample_size,
            )
        case "entire_catalog":
            obj_ids = None
            sql_ids = None
            fact_ids = None
            return await fetch_database_context_alt(
                catalog_con,
                target_con,
                catalog_id,
                sample_size=sample_size,
                obj_ids=obj_ids,
                sql_ids=sql_ids,
                fact_ids=fact_ids,
            )
        case "specific_ids":
            obj_ids = obj_ids or []
            sql_ids = sql_ids or []
            fact_ids = fact_ids or []
            return await fetch_database_context_alt(
                catalog_con,
                target_con,
                catalog_id,
                sample_size=sample_size,
                obj_ids=obj_ids,
                sql_ids=sql_ids,
                fact_ids=fact_ids,
            )
        case _:
            raise RuntimeError(f"unrecognized context mode: {context_mode}")


class IterationLimitExceededException(Exception):
    """Exception raised when iteration limit exceeded."""

    def __init__(self, limit: int):
        self.limit = limit
        self.message = f"Iteration limit exceeded: {limit}"
        super().__init__(self.message)


_SEARCH_TOOL_NAME = "search_for_context"


def _search_tool_definition() -> ToolDefinition:
    return ToolDefinition(
        name=_SEARCH_TOOL_NAME,
        description=(
            "Request additional database context by providing "
            "focused questions for semantic search. Use this when the current "
            "context is insufficient to generate a SQL query confidently. Frame "
            "your questions to target specific tables, relationships, or "
            "attributes needed."
        ),
        parameters_json_schema={
            "type": "object",
            "properties": {
                "questions": {
                    "type": "array",
                    "items": {"type": "string"},
                    "minItems": 1,
                    "description": (
                        "One or more natural language questions for a semantic search "
                        "to find database objects, SQL examples, or facts to provide "
                        "context. Questions should focus on a specific database "
                        "structure, relationship, or calculation. Avoid "
                        "explaining reasoning or context. Provide the search phrase "
                        "only."
                    ),
                }
            },
            "required": ["questions"],
        },
    )


_RECORD_TOOL_NAME = "record_sql_answer"


def _answer_tool_definition() -> ToolDefinition:
    return ToolDefinition(
        name=_RECORD_TOOL_NAME,
        description="Provide a SQL statement that addresses the user's question.",
        parameters_json_schema={
            "type": "object",
            "properties": {
                "sql_statement": {
                    "type": "string",
                    "description": (
                        "A valid SQL statement that accurately "
                        "addresses the user's prompt"
                    ),
                },
                "command_type": {
                    "type": "string",
                    "description": (
                        "Indicate the type of SQL command used in the sql_statement. "
                        "(e.g. 'SELECT', 'INSERT', 'UPDATE', or 'DELETE')"
                    ),
                },
                "relevant_object_ids": {
                    "type": "array",
                    "items": {"type": "integer"},
                    "description": (
                        "Indicate the ids of the database objects that were relevant "
                        "to answering the user's prompt"
                    ),
                },
                "relevant_sql_example_ids": {
                    "type": "array",
                    "items": {"type": "integer"},
                    "description": (
                        "Indicate the ids of the SQL examples that were relevant "
                        "to answering the user's prompt"
                    ),
                },
                "relevant_fact_ids": {
                    "type": "array",
                    "items": {"type": "integer"},
                    "description": (
                        "Indicate the ids of the facts that were relevant to "
                        "answering the user's prompt"
                    ),
                },
            },
            "required": [
                "sql_statement",
                "command_type",
                "relevant_object_ids",
                "relevant_sql_example_ids",
                "relevant_fact_ids",
            ],
        },
    )


async def _increment_obj_usage(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, ids: list[int]
) -> None:
    async with catalog_con.cursor() as cur:
        await cur.execute(
            SQL("""\
            update ai.{} set usage = usage + 1
            where id = any(%s)
        """).format(Identifier(f"semantic_catalog_obj_{catalog_id}")),
            (ids,),
        )


async def _increment_sql_usage(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, ids: list[int]
) -> None:
    async with catalog_con.cursor() as cur:
        await cur.execute(
            SQL("""\
            update ai.{} set usage = usage + 1
            where id = any(%s)
        """).format(Identifier(f"semantic_catalog_sql_{catalog_id}")),
            (ids,),
        )


async def _increment_fact_usage(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, ids: list[int]
) -> None:
    async with catalog_con.cursor() as cur:
        await cur.execute(
            SQL("""\
            update ai.{} set usage = usage + 1
            where id = any(%s)
        """).format(Identifier(f"semantic_catalog_fact_{catalog_id}")),
            (ids,),
        )


async def _freq_used_objects(
    catalog_con: psycopg.AsyncConnection, catalog_id: int, n: int = 20
) -> list[dict[str, Any]]:
    async with catalog_con.cursor(row_factory=dict_row) as cur:
        await cur.execute(
            SQL("""\
            select
              id
            , objtype
            , array_to_string(objnames, '.') as name
            from ai.{}
            where objsubid = 0
            order by usage desc
            limit %s
        """).format(Identifier(f"semantic_catalog_obj_{catalog_id}")),
            (n,),
        )
        return [row for row in await cur.fetchall()]


async def generate_sql(
    catalog_con: psycopg.AsyncConnection,
    target_con: psycopg.AsyncConnection,
    model: KnownModelName | Model,
    catalog_id: int,
    embedding_name: str,
    embedding_config: EmbeddingConfig,
    prompt: str,
    usage: Usage | None = None,
    usage_limits: UsageLimits | None = None,
    model_settings: ModelSettings | None = None,
    iteration_limit: int = 10,
    sample_size: int = 3,
    context_mode: ContextMode = "semantic_search",
    obj_ids: list[int] | None = None,
    sql_ids: list[int] | None = None,
    fact_ids: list[int] | None = None,
) -> GenerateSQLResponse:
    """Generate a SQL statement based on natural language prompt and database context.
    This function uses an AI model to generate a SQL statement that fulfills the user's
    request, drawing on context from the semantic catalog about database objects, SQL
    examples, and facts. The generated SQL is validated against the target database
    to ensure correctness, and refinement iterations are performed if needed.
    The function uses an iterative approach:
    1. Initialize database context based on the specified context mode
    2. Generate a SQL statement using the AI model with the context
    3. Validate the SQL statement against the target database
    4. If invalid, refine the SQL statement with error feedback
    5. Repeat until a valid SQL statement is generated or iteration limit is reached
    Args:
        catalog_con: Connection to the semantic catalog database.
        target_con: Connection to the target database where SQL will be executed.
        model: AI model to use for generating SQL (KnownModelName or Model instance).
        catalog_id: ID of the semantic catalog to use for context.
        embedding_name: Name of the embedding column to use for semantic search.
        embedding_config: Configuration for the embedding model.
        prompt: Natural language prompt describing the desired SQL statement.
        usage: Optional Usage object to track token usage across calls.
        usage_limits: Optional limits on token usage and requests.
        model_settings: Optional settings for the AI model.
        iteration_limit: Maximum number of refinement iterations (default: 5).
        sample_size: Number of sample rows to include for tables/views (default: 3).
        context_mode: Strategy for initializing database context:
            - "semantic_search": Find relevant items semantically (default)
            - "entire_catalog": Include all items from the catalog
            - "specific_ids": Include only items with specified IDs
        obj_ids: Optional list of database object IDs to include (for "specific_ids" mode).
        sql_ids: Optional list of SQL example IDs to include (for "specific_ids" mode).
        fact_ids: Optional list of fact IDs to include (for "specific_ids" mode).
    Returns:
        A GenerateSQLResponse containing:
        - The generated SQL statement
        - The database context used for generation
        - The query plan for the SQL statement
        - The final prompt sent to the model
        - The final response from the model
        - All messages exchanged during generation
        - Usage statistics for the AI model calls
    Raises:
        IterationLimitExceededException: If the iteration limit is reached without
            generating a valid SQL statement.
        RuntimeError: If the semantic catalog is not properly configured.
    Example:
        ```python
        # Generate a SQL statement for a natural language query
        response = await generate_sql(
            catalog_con=catalog_connection,
            target_con=db_connection,
            model="anthropic:claude-3-opus-20240229",
            catalog_id=1,
            embedding_name="openai_embeddings",
            embedding_config=config,
            prompt="Find all orders placed last month with a total value over $1000",
            iteration_limit=3,
        )
        # Use the generated SQL
        print(response.sql_statement)
        ```
    """
    usage = usage or Usage()
    usage_limits = usage_limits or UsageLimits(request_limit=None)
    model_settings = model_settings or ModelSettings()
    iteration_limit = max(int(iteration_limit), 1) or 5

    ctx = await initialize_database_context(
        catalog_con,
        target_con,
        catalog_id,
        embedding_name,
        embedding_config,
        prompt,
        sample_size=sample_size,
        context_mode=context_mode,
        obj_ids=obj_ids,
        sql_ids=sql_ids,
        fact_ids=fact_ids,
    )

    _continue = True
    freq_used_objects: list[dict[str, Any]] = await _freq_used_objects(
        catalog_con, catalog_id
    )
    prior_prompts: list[str] = [prompt]
    answer: str | None = None
    command_type: str | None = None
    pgversion: int | None = await _get_database_version(target_con)
    messages: list[tuple[ModelRequest, ModelResponse]] = []
    user_prompt: str | None = None
    query_plan: dict[str, Any] | None = None
    error: str | None = None
    iteration = 0

    while _continue:
        iteration += 1
        if iteration > iteration_limit:
            raise IterationLimitExceededException(limit=iteration_limit)

        if usage_limits:
            usage_limits.check_before_request(usage)
            if usage_limits.has_token_limits():
                usage_limits.check_tokens(usage)

        system_prompt: str = _template_system_prompt.render(
            pgversion=pgversion, semantic_search_available=(iteration < iteration_limit)
        )

        user_prompt = _template_user_prompt.render(
            ctx=ctx,
            prompt=prompt,
            prior_prompts=prior_prompts,
            freq_used_objects=freq_used_objects,
            error=error,
        )

        request = ModelRequest(
            parts=[
                SystemPromptPart(content=system_prompt),
                UserPromptPart(content=user_prompt),
            ]
        )
        model_response: ModelResponse = await model_request(
            model=model,
            messages=[request],
            model_request_parameters=ModelRequestParameters(
                function_tools=[_search_tool_definition()]
                if iteration < iteration_limit
                else [],
                output_tools=[_answer_tool_definition()],
                allow_text_output=False,
            ),
            model_settings=model_settings,
        )

        messages.append((request, model_response))
        usage = usage + model_response.usage

        for part in model_response.parts:
            match part.part_kind:
                case "tool-call":
                    tool_call_part: ToolCallPart = part
                    args = tool_call_part.args_as_dict()
                    logger.info(f"tool call: {tool_call_part.tool_name}")
                    if tool_call_part.tool_name == _SEARCH_TOOL_NAME:
                        questions: list[str] = (
                            cast(list[str], args.get("questions")) or []
                        )
                        for q in questions:
                            prior_prompts.append(q)
                            logger.info(f"semantic search for '{q}'")
                            ctx = await fetch_database_context(
                                catalog_con,
                                target_con,
                                catalog_id,
                                embedding_name,
                                embedding_config,
                                q,
                                ctx,
                                sample_size=sample_size,
                            )
                    elif tool_call_part.tool_name == _RECORD_TOOL_NAME:
                        answer = args.get("sql_statement")
                        if not answer:
                            logger.error("llm did not provide a sql_statement")
                            continue
                        _continue = False
                        command_type = str(args["command_type"])
                        if command_type.upper() in (
                            "SELECT",
                            "INSERT",
                            "UPDATE",
                            "DELETE",
                            "MERGE",
                            "VALUES",
                        ):
                            logger.info("validating the sql statement")
                            query_plan, error = await validate_sql_statement(
                                target_con, str(answer)
                            )
                            if error:
                                logger.info(
                                    "asking llm to fix the invalid sql statement"
                                )
                                _continue = True
                                continue
                        relevant_object_ids: list[int] | None = cast(
                            list[int] | None, args.get("relevant_object_ids")
                        )
                        relevant_sql_example_ids: list[int] | None = cast(
                            list[int] | None, args.get("relevant_sql_example_ids")
                        )
                        relevant_fact_ids: list[int] | None = cast(
                            list[int] | None, args.get("relevant_fact_ids")
                        )
                        if relevant_object_ids:
                            for irrelevant_id in (
                                ctx.objects.keys() - relevant_object_ids
                            ):
                                ctx.objects.pop(irrelevant_id)
                                ctx.rendered_objects.pop(irrelevant_id)
                            await _increment_obj_usage(
                                catalog_con,
                                catalog_id,
                                [id for id in ctx.objects],
                            )
                        if relevant_sql_example_ids:
                            for irrelevant_id in (
                                ctx.sql_examples.keys() - relevant_sql_example_ids
                            ):
                                ctx.sql_examples.pop(irrelevant_id)
                                ctx.rendered_sql_examples.pop(irrelevant_id)
                            await _increment_sql_usage(
                                catalog_con,
                                catalog_id,
                                [id for id in ctx.sql_examples],
                            )
                        if relevant_fact_ids:
                            for irrelevant_id in ctx.facts.keys() - relevant_fact_ids:
                                ctx.facts.pop(irrelevant_id)
                                ctx.rendered_facts.pop(irrelevant_id)
                            await _increment_fact_usage(
                                catalog_con, catalog_id, [id for id in ctx.facts]
                            )
                    else:
                        logger.error(
                            f"unrecognized tool name: {tool_call_part.tool_name}"
                        )
                case "text":
                    text_part: TextPart = part
                    logger.error(
                        f"unexpected textual model response: {text_part.content}"
                    )
                case _:
                    logger.error(f"unrecognized model response kind: {part.part_kind}")

    assert answer is not None, "answer is unexpectedly missing"
    return GenerateSQLResponse(
        sql_statement=answer,
        context=ctx,
        command_type=command_type or "UNKNOWN",
        query_plan=query_plan or {},
        messages=messages,
        usage=usage,
    )



import asyncio
import datetime
import os
import random
import sys
import traceback
from collections.abc import Sequence
from dataclasses import dataclass

import psycopg
import semver
import structlog
from psycopg.rows import dict_row, namedtuple_row

from .. import __version__
from .embeddings import ApiKeyMixin
from .features import Features
from .vectorizer import Vectorizer
from .worker_tracking import WorkerTracking

if sys.version_info >= (3, 11):
    from builtins import BaseExceptionGroup
else:
    # For Python 3.10 and below, use the backport
    from exceptiongroup import BaseExceptionGroup

logger = structlog.get_logger()


@dataclass
class Version:
    ext_version: str | None
    pgai_lib_version: str | None


class VectorizerNotFoundError(Exception):
    pass


class ApiKeyNotFoundError(Exception):
    pass


def warn_on_old_version(version: Version):
    if version.pgai_lib_version is None:
        if (
            version.ext_version is not None
            and semver.Version.parse(version.ext_version) < "0.10.0"
        ):
            logger.warning(
                "The pgai extension is outdated and can be upgraded. See https://github.com/timescale/pgai/blob/main/docs/vectorizer/migrating-from-extension.md"
            )
    else:
        if semver.Version.parse(__version__) > semver.Version.parse(
            version.pgai_lib_version
        ):
            logger.warning(
                f"The pgai installation in your database is outdated and can be upgraded. Installed version: {version.pgai_lib_version}, latest version: {__version__}."
            )


class Worker:
    def __init__(
        self,
        db_url: str,
        poll_interval: datetime.timedelta = datetime.timedelta(minutes=1),
        once: bool = False,
        vectorizer_ids: Sequence[int] = [],
        exit_on_error: bool | None = None,
        concurrency: int | None = None,
    ):
        self.vectorizer_ids = vectorizer_ids
        self.db_url = db_url
        self.poll_interval = int(poll_interval.total_seconds())
        self.once = once
        self.exit_on_error = exit_on_error
        self.concurrency = concurrency
        self.shutdown_requested = asyncio.Event()

        self.dynamic_mode = len(self.vectorizer_ids) == 0
        if once and exit_on_error is None:
            # once implies exit-on-error
            self.exit_on_error = True

    @staticmethod
    def _get_pgai_version(cur: psycopg.Cursor) -> Version | None:
        cur.execute(
            "select extversion from pg_catalog.pg_extension where extname = 'ai'"
        )
        row = cur.fetchone()
        ext_version = row[0] if row is not None else None

        # todo: think this through more, expecially for Feature Flags
        pgai_lib_version = None
        cur.execute("""
            SELECT EXISTS (
                SELECT 1
                FROM information_schema.tables
                WHERE table_schema = 'ai'
                AND table_name = 'pgai_lib_version'
            )
        """)
        res = cur.fetchone()
        assert res is not None
        table_exists = res[0]
        if table_exists:
            cur.execute("select version from ai.pgai_lib_version where name = 'ai'")
            row = cur.fetchone()
            pgai_lib_version = row[0] if row is not None else None
        return Version(ext_version, pgai_lib_version)

    def _get_vectorizer(self, vectorizer_id: int, features: Features) -> Vectorizer:
        with (
            psycopg.Connection.connect(self.db_url) as con,
            con.cursor(row_factory=dict_row) as cur,
        ):
            cur.execute(
                "select pg_catalog.to_jsonb(v) as vectorizer from ai.vectorizer v where v.id = %s",
                (vectorizer_id,),
            )
            row = cur.fetchone()
            if row is None:
                raise VectorizerNotFoundError(f"vectorizer_id={vectorizer_id}")
            vectorizer = row["vectorizer"]
            embedding = vectorizer["config"]["embedding"]
            vectorizer = Vectorizer.model_validate(vectorizer)

            # The Ollama API doesn't need a key, so `api_key_name` may be unset
            if "api_key_name" in embedding:
                api_key_name = embedding["api_key_name"]
                api_key = os.getenv(api_key_name, None)
                if api_key is not None:
                    logger.debug(f"obtained secret '{api_key_name}' from environment")
                elif features.db_reveal_secrets:
                    cur.execute(
                        "select ai.reveal_secret(%s)",
                        (api_key_name,),
                    )
                    row = cur.fetchone()
                    api_key = row["reveal_secret"] if row is not None else None
                    if api_key is not None:
                        logger.debug(f"obtained secret '{api_key_name}' from database")
                if not api_key:
                    raise ApiKeyNotFoundError(
                        f"api_key_name={api_key_name} vectorizer_id={vectorizer_id}"
                    )
                secrets: dict[str, str | None] = {api_key_name: api_key}
                # The Ollama API doesn't need a key, so doesn't inherit `ApiKeyMixin`
                if isinstance(vectorizer.config.embedding, ApiKeyMixin):
                    vectorizer.config.embedding.set_api_key(secrets)
                else:
                    logger.error(
                        f"cannot set secret value '{api_key_name}' for vectorizer with id: '{vectorizer.id}'"
                    )
            return vectorizer

    async def _handle_error(
        self,
        error_message: str,
        vectorizer_id: int | None,
        worker_tracking: WorkerTracking | None,
    ) -> Exception:
        kwargs = {}
        if vectorizer_id:
            kwargs["vectorizer_id"] = vectorizer_id
        logger.error(error_message, **kwargs)
        if worker_tracking is not None:
            await worker_tracking.save_vectorizer_error(vectorizer_id, error_message)
        if self.exit_on_error:
            if worker_tracking is not None:
                await worker_tracking.force_last_heartbeat_and_stop()
            logger.info("exiting processor due to error")
        return Exception(error_message)

    def _get_vectorizer_ids(
        self, vectorizer_ids: Sequence[int] | None = None
    ) -> list[int]:
        with (
            psycopg.Connection.connect(self.db_url) as con,
            con.cursor(row_factory=namedtuple_row) as cur,
        ):
            valid_vectorizer_ids: list[int] = []
            if vectorizer_ids is None or len(vectorizer_ids) == 0:
                cur.execute("select id from ai.vectorizer")
            else:
                cur.execute(
                    "select id from ai.vectorizer where id = any(%s)",
                    [
                        list(vectorizer_ids),
                    ],
                )
            for row in cur.fetchall():
                valid_vectorizer_ids.append(row[0])
            random.shuffle(valid_vectorizer_ids)
            return valid_vectorizer_ids

    async def request_graceful_shutdown(self):
        """
        Request a graceful shutdown of the processor.
        """
        self.shutdown_requested.set()

    async def run(self) -> Exception | None:
        logger.debug("starting vectorizer worker")

        valid_vectorizer_ids = []
        can_connect = False
        pgai_version = None
        features = None
        worker_tracking = None

        while not self.shutdown_requested.is_set():
            vectorizer_id = None
            try:
                if not can_connect or pgai_version is None:
                    with (
                        psycopg.Connection.connect(self.db_url) as con,
                        con.cursor(row_factory=namedtuple_row) as cur,
                    ):
                        pgai_version = self._get_pgai_version(cur)
                        if pgai_version is None or (
                            pgai_version.ext_version is None
                            and pgai_version.pgai_lib_version is None
                        ):
                            err_msg = "pgai is not installed in the database"
                            exception = await self._handle_error(
                                err_msg, None, worker_tracking
                            )
                            if self.exit_on_error:
                                return exception
                        else:
                            warn_on_old_version(pgai_version)
                            features = Features.from_db(cur)
                            worker_tracking = WorkerTracking(
                                self.db_url, self.poll_interval, features, __version__
                            )
                            await worker_tracking.start()
                            can_connect = True

                if can_connect and features is not None and worker_tracking is not None:
                    if not self.dynamic_mode and len(valid_vectorizer_ids) != len(
                        self.vectorizer_ids
                    ):
                        valid_vectorizer_ids = self._get_vectorizer_ids(
                            self.vectorizer_ids,
                        )
                        if len(valid_vectorizer_ids) != len(self.vectorizer_ids):
                            err_msg = f"invalid vectorizers, wanted: {list(self.vectorizer_ids)}, got: {valid_vectorizer_ids}"
                            exception = await self._handle_error(
                                err_msg,
                                None,
                                worker_tracking,
                            )
                            if self.exit_on_error:
                                return exception
                    else:
                        valid_vectorizer_ids = self._get_vectorizer_ids(
                            self.vectorizer_ids,
                        )
                        if len(valid_vectorizer_ids) == 0:
                            logger.warning("no vectorizers found")

                    for vectorizer_id in valid_vectorizer_ids:
                        try:
                            vectorizer = self._get_vectorizer(vectorizer_id, features)
                        except (VectorizerNotFoundError, ApiKeyNotFoundError) as e:
                            err_msg = f"error getting vectorizer: {type(e).__name__}: {str(e)}"
                            exception = await self._handle_error(
                                err_msg, vectorizer_id, worker_tracking
                            )
                            if self.exit_on_error:
                                return exception
                            break

                        logger.info("running vectorizer", vectorizer_id=vectorizer_id)

                        def should_continue(_: int, __: int) -> bool:
                            return not self.shutdown_requested.is_set()

                        await vectorizer.run(
                            db_url=self.db_url,
                            features=features,
                            worker_tracking=worker_tracking,
                            concurrency=self.concurrency,
                            should_continue_processing_hook=should_continue,
                        )
            except psycopg.OperationalError as e:
                if "connection failed" in str(e):
                    err_msg = f"unable to connect to database: {str(e)}"
                else:
                    err_msg = f"unexpected error: {str(e)}"
                exception = await self._handle_error(
                    err_msg, vectorizer_id, worker_tracking
                )
                if self.exit_on_error:
                    return exception
            except BaseExceptionGroup as e:  # type: ignore
                # catch any exceptions, log them, and keep on going
                for exception in e.exceptions:  # type: ignore
                    error_msg = str(exception)  # type: ignore
                    logger.error(error_msg, vectorizer_id=vectorizer_id)
                for exception_line in traceback.format_exception(e):  # type: ignore
                    for line in exception_line.rstrip().split("\n"):
                        logger.debug(line)
                err_msg = f"unexpected error: {str(e)}"  # type: ignore
                exception = await self._handle_error(
                    err_msg, vectorizer_id, worker_tracking
                )
                if self.exit_on_error:
                    return exception
            except Exception as e:
                # catch any exceptions, log them, and keep on going
                for exception_line in traceback.format_exception(e):
                    for line in exception_line.rstrip().split("\n"):
                        logger.debug(line)
                err_msg = f"unexpected error: {str(e)}"
                exception = await self._handle_error(
                    err_msg, vectorizer_id, worker_tracking
                )
                if self.exit_on_error:
                    return exception

            if self.once:
                if worker_tracking is not None:
                    await worker_tracking.force_last_heartbeat_and_stop()
                logger.info("once mode, exiting...")
                return None

            poll_interval_str = datetime.timedelta(seconds=self.poll_interval)
            logger.info(f"sleeping for {poll_interval_str} before polling for new work")
            try:
                await asyncio.wait_for(
                    self.shutdown_requested.wait(), timeout=self.poll_interval
                )
                # shutdown event was set, the loop should exit
                logger.info("got a graceful shutdown request")
            except asyncio.TimeoutError:
                # timeout means the sleep completed
                pass

        if worker_tracking is not None:
            await worker_tracking.force_last_heartbeat_and_stop()
        logger.info("exiting processor.run()")
        return None



from typing import Literal

from pydantic import BaseModel


class ObjectDescription(BaseModel):
    """Model representing a description of a database object.

    This class represents the metadata and description of database objects such as
    tables, views, functions, etc. It includes identifiers that can be used to
    locate the object in the database, as well as a textual description.

    Attributes:
        id: Semantic catalog ID (default: -1 if not assigned).
        classid: PostgreSQL object class ID (for pg_class, pg_proc, etc.).
        objid: PostgreSQL object ID within its class.
        objsubid: PostgreSQL sub-object ID (e.g., column number).
        objtype: Type of object (table, view, function, etc.).
        objnames: Object name components (schema, name, etc.).
        objargs: Object argument types for procedures/functions.
        description: Textual description of the object.
    """

    id: int = -1
    classid: int
    objid: int
    objsubid: int
    objtype: str
    objnames: list[str]
    objargs: list[str]
    description: str


class Column(BaseModel):
    """Model representing a database column.

    This class represents a column in a database table or view, including its
    data type, constraints, and description.

    Attributes:
        classid: PostgreSQL object class ID for the parent table/view.
        objid: PostgreSQL object ID of the parent table/view.
        objsubid: PostgreSQL sub-object ID (column number).
        name: Column name.
        type: PostgreSQL data type of the column.
        is_not_null: Boolean indicating if the column has a NOT NULL constraint.
        default_value: Default value expression for the column (if any).
        description: Object description containing metadata and textual description.
    """

    classid: int
    objid: int
    objsubid: int
    name: str
    type: str
    is_not_null: bool
    default_value: str | None = None
    description: ObjectDescription | None = None


class Dimension(BaseModel):
    """Model representing a dimension in a TimescaleDB hypertable.

    This class represents a partitioning dimension in a TimescaleDB hypertable,
    which can be time-based or space-based.

    Attributes:
        column_name: Name of the column used for partitioning.
        dimension_builder: Type of partitioning strategy (by_range, by_hash).
        partition_func: Custom partitioning function (if any).
        partition_interval: Time interval for time-based partitioning.
        number_partitions: Number of partitions for hash partitioning.
    """

    column_name: str
    dimension_builder: str
    partition_func: str | None = None
    partition_interval: str | None = None
    number_partitions: int | None = None


class View(BaseModel):
    """Model representing a database view.

    This class represents a view in a database, which can be a regular view,
    materialized view, or a TimescaleDB continuous aggregate view.

    Attributes:
        id: Semantic catalog ID (default: -1 if not assigned).
        classid: PostgreSQL object class ID.
        objid: PostgreSQL object ID.
        schema_name: Schema name where the view is defined.
        view_name: Name of the view.
        is_materialized: Boolean indicating if the view is materialized.
        is_continuous_aggregate: Boolean indicating if the view is a TimescaleDB continuous aggregate.
        columns: List of columns in the view.
        definition: SQL definition of the view.
        description: Object description containing metadata and textual description.
        sample: Sample data from the view (if available).
    """

    id: int = -1
    classid: int
    objid: int
    schema_name: str
    view_name: str
    is_materialized: bool
    is_continuous_aggregate: bool
    columns: list[Column] | None = None
    definition: str
    description: ObjectDescription | None = None
    sample: str | None = None


class Procedure(BaseModel):
    """Model representing a database procedure, function, or aggregate.

    This class represents a callable database object, which can be a procedure,
    function, or aggregate function.

    Attributes:
        id: Semantic catalog ID (default: -1 if not assigned).
        classid: PostgreSQL object class ID.
        objid: PostgreSQL object ID.
        schema_name: Schema name where the procedure is defined.
        proc_name: Name of the procedure.
        kind: Type of procedure ("procedure", "function", or "aggregate").
        identity_args: String representation of the argument types for identification.
        definition: SQL definition of the procedure.
        objargs: List of argument types as strings.
        description: Object description containing metadata and textual description.
    """

    id: int = -1
    classid: int
    objid: int
    schema_name: str
    proc_name: str
    kind: Literal["procedure", "function", "aggregate"]
    identity_args: str
    definition: str
    objargs: list[str] = []
    description: ObjectDescription | None = None


class Constraint(BaseModel):
    """Model representing a database constraint.

    This class represents a constraint in a database table, such as primary key,
    foreign key, unique, or check constraints.

    Attributes:
        name: Name of the constraint.
        definition: SQL definition of the constraint.
    """

    name: str
    definition: str


class Index(BaseModel):
    """Model representing a database index.

    This class represents an index on a database table.

    Attributes:
        name: Name of the index.
        definition: SQL definition of the index.
    """

    name: str
    definition: str


class Table(BaseModel):
    """Model representing a database table.

    This class represents a table in a database, including its columns, constraints,
    indexes, and potentially TimescaleDB hypertable dimensions.

    Attributes:
        id: Semantic catalog ID (default: -1 if not assigned).
        classid: PostgreSQL object class ID.
        objid: PostgreSQL object ID.
        schema_name: Schema name where the table is defined.
        table_name: Name of the table.
        persistence: Table persistence type ("temporary" or "unlogged") if applicable.
        columns: List of columns in the table.
        constraints: List of constraints on the table.
        indexes: List of indexes on the table.
        dimensions: List of TimescaleDB dimensions if the table is a hypertable.
        description: Object description containing metadata and textual description.
        sample: Sample data from the table (if available).
    """

    id: int = -1
    classid: int
    objid: int
    schema_name: str
    table_name: str
    persistence: Literal["temporary", "unlogged"] | None = None
    columns: list[Column] | None = None
    constraints: list[Constraint] | None = None
    indexes: list[Index] | None = None
    dimensions: list[Dimension] | None = None
    description: ObjectDescription | None = None
    sample: str | None = None


class SQLExample(BaseModel):
    """Model representing an example SQL query.

    This class represents an example SQL query with a description of what it does.
    These examples can be used to help users understand how to query the database.

    Attributes:
        id: Semantic catalog ID (default: -1 if not assigned).
        sql: The SQL query text.
        description: Description of what the SQL query does.
    """

    id: int = -1
    sql: str
    description: str


class Fact(BaseModel):
    """Model representing a fact about the database or domain.

    This class represents a descriptive fact about the database, its schema,
    or the business domain it models. These facts can provide context and information
    that isn't captured in the specific database object descriptions.

    Attributes:
        id: Semantic catalog ID (default: -1 if not assigned).
        description: The text of the fact.
    """

    id: int = -1
    description: str



import re
from collections.abc import AsyncGenerator
from functools import cached_property
from typing import TYPE_CHECKING, Literal

import ijson  # type: ignore
from pydantic import BaseModel
from typing_extensions import override

if TYPE_CHECKING:
    import openai
    import tiktoken
    from openai import AsyncAPIResponse, resources, types


from ..embeddings import (
    ApiKeyMixin,
    BaseURLMixin,
    Embedder,
    EmbeddingResponse,
    EmbeddingVector,
    Usage,
    logger,
)

TOKEN_CONTEXT_LENGTH_ERROR = "chunk exceeds model context length"

EMBEDDING_MODEL_CONTEXT_LENGTH = {
    "text-embedding-ada-002": 8191,
    "text-embedding-3-small": 8191,
    "text-embedding-3-large": 8191,
}

# ijson reads the input in chunks of buf_size. Since we are wrapping our
# iterator over the read implementation that ijson requires, we explicitly set
# the value when calling ijson methods, and use the same value to iterate over
# the openAI stream response.
#
# The value of 64KB is the default from ijson.
RESPONSE_READ_BUF_SIZE = 64 * 1024

openai_token_length_regex = re.compile(
    r"This model's maximum context length is (\d+) tokens"
)


class ResponseWithRead:
    def __init__(self, response: "AsyncAPIResponse[types.CreateEmbeddingResponse]"):
        self.iter = response.iter_bytes(RESPONSE_READ_BUF_SIZE)

    async def read(self, n: int) -> bytes:
        if n == 0:
            return b""

        return await anext(self.iter)


class OpenAI(ApiKeyMixin, BaseURLMixin, BaseModel, Embedder):
    """
    Embedder that uses OpenAI's API to embed documents into vector representations.

    Attributes:
        implementation (Literal["openai"]): The literal identifier for this
            implementation.
        model (str): The name of the OpenAI model used for embeddings.
        dimensions (int | None): Optional dimensions for the embeddings.
        user (str | None): Optional user identifier for OpenAI API usage.
    """

    implementation: Literal["openai"]
    model: str
    dimensions: int | None = None
    user: str | None = None

    @cached_property
    def _openai_dimensions(self) -> "int | openai.NotGiven":
        # Note: deferred import to avoid import overhead
        import openai

        if self.model == "text-embedding-ada-002":
            if self.dimensions != 1536:
                raise ValueError("dimensions must be 1536 for text-embedding-ada-002")
            return openai.NOT_GIVEN
        return self.dimensions if self.dimensions is not None else openai.NOT_GIVEN

    @cached_property
    def _openai_user(self) -> "str | openai.NotGiven":
        # Note: deferred import to avoid import overhead
        import openai

        return self.user if self.user is not None else openai.NOT_GIVEN

    @cached_property
    def _embedder(self) -> "resources.AsyncEmbeddingsWithStreamingResponse":
        import openai

        return openai.AsyncOpenAI(
            base_url=self.base_url, api_key=self._api_key, max_retries=3
        ).embeddings.with_streaming_response

    @override
    def _max_chunks_per_batch(self) -> int:
        return 2048

    @override
    def _max_tokens_per_batch(self) -> int:
        return 300_000

    @override
    async def call_embed_api(self, documents: list[str]) -> EmbeddingResponse:
        embeddings: list[list[float]] = []
        current_embedding: list[float] = []
        total_tokens = 0
        prompt_tokens = 0
        async with self._embedder.create(
            input=documents,
            model=self.model,
            dimensions=self._openai_dimensions,
            user=self._openai_user,
            encoding_format="float",
        ) as streaming_response:
            # We could simplify by using ijson.item_async:
            #
            # async for value in ijson.items(
            #    ResponseWithRead(raw_response), "data.item.embedding", use_float=True
            # ):
            #     embeddings.append(value)
            #
            # The problem is that `ijson.items` accepts only one prefix, and we
            # need to read data from 2, `data.item.embedding` and `usage`.
            # There's a WIP PR to support this use case:
            # https://github.com/ICRAR/ijson/pull/127
            async for prefix, event, value in ijson.parse_async(
                ResponseWithRead(streaming_response),
                use_float=True,
                buf_size=RESPONSE_READ_BUF_SIZE,
            ):
                if prefix == "data.item.embedding" and event == "start_array":
                    current_embedding = []
                if prefix == "data.item.embedding" and event == "end_array":
                    embeddings.append(current_embedding)
                elif prefix == "data.item.embedding.item" and event == "number":
                    current_embedding.append(value)
                elif prefix == "usage.prompt_tokens" and event == "number":
                    prompt_tokens = value
                elif prefix == "usage.total_tokens" and event == "number":
                    total_tokens = value

        return EmbeddingResponse(
            embeddings=embeddings, usage=Usage(prompt_tokens, total_tokens)
        )

    def _estimate_token_length(self, document: str) -> float:
        """
        Estimates token count based on UTF-8 byte length.
        """

        total_estimated_tokens = 0
        for char in document:
            byte_length = len(char.encode("utf-8"))
            total_estimated_tokens += byte_length * 0.25  # 0.25 tokens per byte

        return total_estimated_tokens

    @override
    async def embed(
        self, documents: list[str]
    ) -> AsyncGenerator[list[EmbeddingVector], None]:
        """
        Embeds a list of documents into vectors using OpenAI's embeddings API.
        The documents are first encoded into tokens before being embedded and
        truncated to not exceed OpenAI's context window.

        Args:
            documents (list[str]): A list of documents to be embedded.

        Returns:
            AsyncGenerator[list[EmbeddingVector], None]: The embeddings for
            each document.
        """
        await logger.adebug(f"Chunks produced: {len(documents)}")
        encoder = self._encoder
        context_length = self._context_length
        if encoder is not None and context_length is not None:
            # truncate all documents before submitting them to the API
            for i, document in enumerate(documents):
                tokenized = encoder.encode(document)
                tokenized_length = len(tokenized)
                if tokenized_length > context_length:
                    await logger.awarning(
                        f"chunk truncated from {len(tokenized)} to {context_length} tokens"
                    )
                    documents[i] = encoder.decode(tokenized[:context_length])
        # OpenAIs per batch token limit is using a token estimator instead of actual tokens
        # So we are reproducing their token counts
        token_counts = [self._estimate_token_length(document) for document in documents]
        async for embeddings in self.batch_chunks_and_embed(documents, token_counts):
            yield embeddings

    @cached_property
    def _encoder(self) -> "tiktoken.Encoding | None":
        # Note: deferred import to avoid import overhead
        import tiktoken

        try:
            encoder = tiktoken.encoding_for_model(self.model)
        except KeyError:
            logger.warning(f"Tokenizer for the model {self.model} not found.")
            return None
        return encoder

    @cached_property
    def _context_length(self) -> int | None:
        return EMBEDDING_MODEL_CONTEXT_LENGTH.get(self.model, None)



"""Templates module for the semantic catalog.

This module provides a Jinja2 environment and templates for rendering database objects
(tables, views, procedures), SQL examples, facts, and prompts for SQL generation.
The templates are used to format the catalog items for display, export, and AI
interaction.
"""

from pathlib import Path

from jinja2 import Environment, FileSystemLoader

# Create a Jinja2 environment with the templates directory as the loader path
env = Environment(loader=FileSystemLoader(Path(__file__).parent.joinpath("templates")))



# ruff: noqa: E402

"""Semantic Catalog for PostgreSQL databases.

This package provides functionality for creating and managing semantic catalogs
that store metadata about database objects along with natural language descriptions
and vector embeddings. The semantic catalog enables:

- Document database schemas with AI-generated natural language descriptions
- Search database objects using natural language queries
- Generate SQL statements based on natural language prompts
- Manage database documentation

Main components include:
- SemanticCatalog: Main class for interacting with a semantic catalog
- Management functions: create, from_id, from_name, list_semantic_catalogs
- Vectorization: Create embeddings for database objects to enable semantic search
- SQL generation: Generate SQL statements based on natural language queries
"""

import sys
from importlib.util import find_spec

rich = find_spec("rich")
if rich is None:
    print(
        "semantic-catalog extra is not installed, please install it with `pip install pgai[semantic-catalog]`",  # noqa: E501
        file=sys.stderr,
    )
    sys.exit(1)

from .semantic_catalog import (
    CatalogConnection,
    SemanticCatalog,
    TargetConnection,
    create,
    from_id,
    from_name,
    list_semantic_catalogs,
)

__all__ = [
    "list_semantic_catalogs",
    "from_name",
    "from_id",
    "create",
    "SemanticCatalog",
    "TargetConnection",
    "CatalogConnection",
]



import asyncio
import json
import os
import sys
import threading
import time
from collections.abc import AsyncGenerator, Callable, Sequence
from functools import cache, cached_property, partial
from itertools import islice
from typing import Any, TypeAlias, TypeVar
from uuid import UUID

import psycopg
import structlog
from ddtrace.trace import tracer
from pgvector.psycopg import register_vector_async  # type: ignore
from psycopg import AsyncConnection, sql
from psycopg.rows import dict_row
from psycopg.types.json import Jsonb, set_json_dumps
from pydantic import BaseModel, Field, model_validator
from pydantic_core._pydantic_core import ArgsKwargs
from typing_extensions import override

from .chunking import (
    LangChainCharacterTextSplitter,
    LangChainRecursiveCharacterTextSplitter,
    NoneChunker,
)
from .destination import ColumnDestination, TableDestination
from .embedders import LiteLLM, Ollama, OpenAI, VoyageAI
from .features import Features
from .formatting import ChunkValue, PythonTemplate
from .loading import ColumnLoading, LoadingError, UriLoading
from .migrations import apply_migrations
from .parsing import ParsingAuto, ParsingNone, ParsingPyMuPDF
from .processing import ProcessingDefault
from .worker_tracking import WorkerTracking

logger = structlog.get_logger()

VectorizerErrorRecord: TypeAlias = tuple[int, str, Jsonb]
EmbeddingRecord: TypeAlias = list[Any]
SourceRow: TypeAlias = dict[str, Any]

DEFAULT_CONCURRENCY = 1
DEFAULT_VECTORIZER_ERRORS_TABLE = "_vectorizer_errors"

VECTORIZER_FAILED = "vectorizer failed with unexpected error"

if sys.version_info >= (3, 11):
    from builtins import BaseExceptionGroup
else:
    # For Python 3.10 and below, use the backport
    from exceptiongroup import BaseExceptionGroup


class EmbeddingProviderError(Exception):
    """
    Raised when an embedding provider API request fails.
    """

    msg = "embedding provider failed"


class PkAtt(BaseModel):
    """
    Represents an attribute of a primary key.

    Attributes:
        attname (str): The name of the attribute (column).
    """

    attname: str
    pknum: int
    attnum: int


class Config(BaseModel):
    """
    Holds the configuration for the vectorizer including embedding, processing,
    chunking, and formatting.

    Attributes:
        version: The version of the configuration.
        embedding: The embedding's provider configuration.
        processing: Processing settings such as batch size and concurrency.
        chunking: The chunking strategy.
        formatting: Formatting strategy to apply to the chunks.
    """

    version: str
    # Set in the migrations if the configuration is migrated to a newer version
    original_version: str | None = None
    loading: ColumnLoading | UriLoading
    embedding: OpenAI | Ollama | VoyageAI | LiteLLM
    processing: ProcessingDefault
    destination: TableDestination | ColumnDestination = Field(
        ..., discriminator="implementation"
    )
    chunking: (
        LangChainCharacterTextSplitter
        | LangChainRecursiveCharacterTextSplitter
        | NoneChunker
    ) = Field(..., discriminator="implementation")
    formatting: PythonTemplate | ChunkValue = Field(..., discriminator="implementation")
    parsing: ParsingNone | ParsingAuto | ParsingPyMuPDF = Field(
        default_factory=lambda: ParsingAuto(implementation="auto"),
        discriminator="implementation",
    )


class Vectorizer(BaseModel):
    """
    Represents a vectorizer configuration that processes data from a source
    table to generate embeddings.

    Attributes:
        id (int): The unique identifier of the vectorizer.
        config (Config): The configuration object for the vectorizer.
        queue_table (str): The name of the queue table.
        queue_schema (str): The schema where the queue table is located.
        source_schema (str): The schema of the source table.
        source_table (str): The source table where the data comes from.
        target_schema (str): The schema of the target table where embeddings are saved.
        target_table (str): The target table where embeddings are saved.
        source_pk (list[PkAtt]): List of primary key attributes from the source table.
        errors_schema (str): The schema where the error log is saved. Default is "ai".
        errors_table (str): The table where errors are logged.
    """

    id: int
    config: Config
    queue_schema: str
    queue_table: str
    source_schema: str
    source_table: str
    source_pk: list[PkAtt]
    queue_failed_table: str | None = None
    errors_schema: str = "ai"
    errors_table: str = DEFAULT_VECTORIZER_ERRORS_TABLE
    schema_: str = Field(alias="schema", default="ai")
    table: str = "vectorizer"

    async def run(
        self,
        db_url: str,
        features: Features,
        worker_tracking: WorkerTracking,
        concurrency: int | None = None,
        should_continue_processing_hook: None | Callable[[int, int], bool] = None,
    ) -> int:
        """Run this vectorizer with the specified configuration using Worker instances

        Args:
            db_url: Database connection URL
            features: Features from database
            worker_tracking: Tracking instance for worker stats
            concurrency: Number of concurrent workers
                (overrides vectorizer config if provided)
            should_continue_processing_hook: Optional callback to
                control processing flow

        Returns:
            Number of items processed
        """

        if (
            self.errors_table == DEFAULT_VECTORIZER_ERRORS_TABLE
            and not features.has_vectorizer_errors_view
        ):
            self.errors_table = "vectorizer_errors"

        concurrency = concurrency or self.config.processing.concurrency
        tasks = [
            asyncio.create_task(
                Executor(
                    db_url,
                    self,
                    features,
                    worker_tracking,
                    should_continue_processing_hook,
                ).run()
            )
            for _ in range(concurrency)
        ]

        results = await asyncio.gather(*tasks, return_exceptions=True)

        # raise any exceptions, but only after all tasks have completed
        items: int = 0
        exceptions: list[BaseException] = []
        for result in results:
            if isinstance(result, BaseException):
                # report all exceptions. Doesn't log, it just stores in the DB.
                await worker_tracking.save_vectorizer_error(self.id, str(result))
                exceptions.append(result)
            else:
                items += result

        logger.info(
            "finished processing vectorizer", items=items, vectorizer_id=self.id
        )

        if len(exceptions) > 0:
            raise BaseExceptionGroup("vectorizer finished with errors", exceptions)

        return items

    @model_validator(mode="before")
    @classmethod
    def migrate_config_to_new_version(cls, data: Any) -> Any:
        if not data:
            return data
        if isinstance(data, ArgsKwargs) and data.kwargs is not None:
            return apply_migrations(data.kwargs)
        if isinstance(data, dict) and all(isinstance(key, str) for key in data):  # type: ignore[reportUnknownVariableType]
            return apply_migrations(data)  # type: ignore[arg-type]

        logger.warning("Unable to migrate configuration: raw data type is unknown")
        return data  # type: ignore[reportUnknownVariableType]


class VectorizerQueryBuilder:
    """
    A query builder class for generating SQL queries related to the vectorizer
    operations.

    Attributes:
        vectorizer (Vectorizer): The vectorizer for which queries are built.
    """

    def __init__(self, vectorizer: Vectorizer):
        self.vectorizer = vectorizer

    @property
    def pk_fields_sql(self) -> sql.Composed:
        """Generates the SQL expression for a comma separated list of the
        attributes of a primary key. For example, if the primary key has 2
        fields (author, title), it will return a sql.Composed object that
        represents the SQL expression "author, title". The columns will be
        listed in the order they appear in the table NOT the primary key.
        """
        return sql.SQL(" ,").join(
            [
                sql.Identifier(a.attname)
                for a in sorted(self.vectorizer.source_pk, key=lambda pk: pk.attnum)
            ]
        )

    @cached_property
    def pk_attnames(self) -> list[str]:
        """
        Returns a list of primary key attribute names. For example, if the
        primary key has 2 fields (author, title), it will return ["author", "title"].
        The columns will be listed in the order they appear in the table NOT the
        primary key.
        """
        return [
            a.attname
            for a in sorted(self.vectorizer.source_pk, key=lambda pk: pk.attnum)
        ]

    @property
    def pk_fields(self) -> list[sql.Identifier]:
        """
        Returns the SQL identifiers for primary key fields.
        The columns will be listed in the order they appear in the table NOT the
        primary key.
        """
        return [
            sql.Identifier(a.attname)
            for a in sorted(self.vectorizer.source_pk, key=lambda pk: pk.attnum)
        ]

    @cache  # noqa: B019
    def target_table_ident(self, destination: TableDestination) -> sql.Identifier:
        """
        Returns the SQL identifier for the fully qualified name of the target table.
        """
        return sql.Identifier(
            destination.target_schema,
            destination.target_table,
        )

    @property
    def source_table_ident(self) -> sql.Identifier:
        """
        Returns the SQL identifier for the fully qualified name of the source table.
        """
        return sql.Identifier(
            self.vectorizer.source_schema, self.vectorizer.source_table
        )

    @property
    def errors_table_ident(self) -> sql.Identifier:
        """
        Returns the SQL identifier for the fully qualified name of the errors table.
        """
        return sql.Identifier(
            self.vectorizer.errors_schema, self.vectorizer.errors_table
        )

    @property
    def queue_table_ident(self) -> sql.Identifier:
        """
        Returns the SQL identifier for the fully qualified name of the queue table.
        """
        return sql.Identifier(self.vectorizer.queue_schema, self.vectorizer.queue_table)

    @property
    def vectorizer_table_ident(self) -> sql.Identifier:
        """
        Returns the SQL identifier for the fully qualified name of the
        vectorizer table.
        """
        return sql.Identifier(self.vectorizer.schema_, self.vectorizer.table)

    @cached_property
    def fetch_work_query(self) -> sql.Composed:
        """
        Generates the SQL query to fetch work items from the queue table.

        The query is safe to run concurrently from multiple workers. It handles
        duplicate work items by allowing only one instance of the duplicates to
        be proccessed at a time.

        For a thorough explanation of the query, see:

        https://www.timescale.com/blog/how-we-designed-a-resilient-vector-embedding-creation-system-for-postgresql-data/#process-the-work-queue

        The main takeaways are:

        > ... the system retrieves a specified number of entries from the work
        queue, determined by the batch queue size parameter. A FOR UPDATE lock
        is taken to ensure that concurrently executing scripts don’t try
        processing the same queue items. The SKIP LOCKED directive ensures that
        if any entry is currently being handled by another script, the system
        will skip it instead of waiting, avoiding unnecessary delays.

        > Due to the possibility of duplicate entries for the same blog_id
        within the work-queue table, simply locking said table is insufficient
        ... A Postgres advisory lock, prefixed with the table identifier to
        avoid potential overlaps with other such locks, is employed. The try
        variant, analogous to the earlier application of SKIP LOCKED, ensures
        the system avoids waiting on locks. The inclusion of the ORDER BY
        blog_id clause helps prevent potential deadlocks...

        The only differece, between the blog and this query, is that we handle
        composite primary keys.
        """
        return sql.SQL("""
                WITH selected_rows AS (
                    SELECT {pk_fields}
                    FROM {queue_table}
                    LIMIT %s
                    FOR UPDATE SKIP LOCKED
                ),
                locked_items AS (
                    SELECT
                        {pk_fields},
                        pg_try_advisory_xact_lock(
                            %s,
                            hashtext(concat_ws('|', {lock_fields}))
                        ) AS locked
                    FROM (
                        SELECT DISTINCT {pk_fields}
                        FROM selected_rows
                        ORDER BY {pk_fields}
                    ) as ids
                ),
                deleted_rows AS (
                    DELETE FROM {queue_table} AS w
                    USING locked_items AS l
                    WHERE l.locked = true
                    AND {delete_join_predicates}
                )
                SELECT s.*
                FROM locked_items l
                LEFT JOIN LATERAL ( -- NOTE: lateral join forces runtime chunk exclusion
                    SELECT *
                    FROM {source_schema}.{source_table} s
                    WHERE {lateral_join_predicates}
                    LIMIT 1
                ) AS s ON true
                WHERE l.locked = true
                ORDER BY {pk_fields}
                        """).format(
            pk_fields=self.pk_fields_sql,
            queue_table=sql.Identifier(
                self.vectorizer.queue_schema, self.vectorizer.queue_table
            ),
            lock_fields=sql.SQL(" ,").join(
                [
                    xs
                    for x in self.vectorizer.source_pk
                    for xs in [
                        sql.Literal(x.attname),
                        sql.Identifier(x.attname),
                    ]
                ]
            ),
            delete_join_predicates=sql.SQL(" AND ").join(
                [
                    sql.SQL("w.{} = l.{}").format(
                        sql.Identifier(x.attname),
                        sql.Identifier(x.attname),
                    )
                    for x in self.vectorizer.source_pk
                ]
            ),
            source_schema=sql.Identifier(self.vectorizer.source_schema),
            source_table=sql.Identifier(self.vectorizer.source_table),
            lateral_join_predicates=sql.SQL(" AND ").join(
                [
                    sql.SQL("l.{} = s.{}").format(
                        sql.Identifier(x.attname),
                        sql.Identifier(x.attname),
                    )
                    for x in self.vectorizer.source_pk
                ]
            ),
        )

    @cached_property
    def fetch_work_query_with_retries(self) -> sql.Composed:
        """
        Note that this is an updated version of the original query that includes
        the loading_retries column.
        Generates the SQL query to fetch work items from the queue table.

        The query is safe to run concurrently from multiple workers. It handles
        duplicate work items by allowing only one instance of the duplicates to
        be proccessed at a time.

        For a thorough explanation of the query, see:

        https://www.timescale.com/blog/how-we-designed-a-resilient-vector-embedding-creation-system-for-postgresql-data/#process-the-work-queue

        The main takeaways are:

        > ... the system retrieves a specified number of entries from the work
        queue, determined by the batch queue size parameter. A FOR UPDATE lock
        is taken to ensure that concurrently executing scripts don't try
        processing the same queue items. The SKIP LOCKED directive ensures that
        if any entry is currently being handled by another script, the system
        will skip it instead of waiting, avoiding unnecessary delays.

        > Due to the possibility of duplicate entries for the same blog_id
        within the work-queue table, simply locking said table is insufficient
        ... A Postgres advisory lock, prefixed with the table identifier to
        avoid potential overlaps with other such locks, is employed. The try
        variant, analogous to the earlier application of SKIP LOCKED, ensures
        the system avoids waiting on locks. The inclusion of the ORDER BY
        blog_id clause helps prevent potential deadlocks...

        The only difference, between the blog and this query, is that we handle
        composite primary keys.
        """
        return sql.SQL("""
                WITH selected_rows AS (
                    SELECT {pk_fields}, {loading_retries}
                    FROM {queue_table}
                    WHERE loading_retry_after is null or loading_retry_after < now()
                    LIMIT %s
                    FOR UPDATE SKIP LOCKED
                ),
                locked_items AS (
                    SELECT
                        {pk_fields}, {loading_retries},
                        pg_try_advisory_xact_lock(
                            %s,
                            hashtext(concat_ws('|', {lock_fields}))
                        ) AS locked
                    FROM (
                        SELECT DISTINCT {pk_fields}, {loading_retries}
                        FROM selected_rows
                        ORDER BY {pk_fields}
                    ) as ids
                ),
                deleted_rows AS (
                    DELETE FROM {queue_table} AS w
                    USING locked_items AS l
                    WHERE l.locked = true
                    AND {delete_join_predicates}
                )
                SELECT s.*, {loading_retries}
                FROM locked_items l
                LEFT JOIN LATERAL ( -- NOTE: lateral join forces runtime chunk exclusion
                    SELECT *
                    FROM {source_schema}.{source_table} s
                    WHERE {lateral_join_predicates}
                    LIMIT 1
                ) AS s ON true
                WHERE l.locked = true
                ORDER BY {pk_fields}
                        """).format(
            pk_fields=self.pk_fields_sql,
            loading_retries=sql.Identifier("loading_retries"),
            queue_table=sql.Identifier(
                self.vectorizer.queue_schema, self.vectorizer.queue_table
            ),
            lock_fields=sql.SQL(" ,").join(
                [
                    xs
                    for x in self.vectorizer.source_pk
                    for xs in [
                        sql.Literal(x.attname),
                        sql.Identifier(x.attname),
                    ]
                ]
            ),
            delete_join_predicates=sql.SQL(" AND ").join(
                [
                    sql.SQL("w.{} = l.{}").format(
                        sql.Identifier(x.attname),
                        sql.Identifier(x.attname),
                    )
                    for x in self.vectorizer.source_pk
                ]
            ),
            source_schema=sql.Identifier(self.vectorizer.source_schema),
            source_table=sql.Identifier(self.vectorizer.source_table),
            lateral_join_predicates=sql.SQL(" AND ").join(
                [
                    sql.SQL("l.{} = s.{}").format(
                        sql.Identifier(x.attname),
                        sql.Identifier(x.attname),
                    )
                    for x in self.vectorizer.source_pk
                ]
            ),
        )

    @cached_property
    def fetch_queue_table_oid_query(self) -> sql.Composed:
        return sql.SQL("SELECT to_regclass('{}')::oid").format(
            sql.Identifier(self.vectorizer.queue_schema, self.vectorizer.queue_table)
        )

    def delete_embeddings_query(
        self, items_count: int, destination: TableDestination
    ) -> sql.Composed:
        return sql.SQL("DELETE FROM {} WHERE ({}) IN ({})").format(
            self.target_table_ident(destination),  # type: ignore
            self.pk_fields_sql,
            self._pks_placeholders_tuples(items_count),
        )

    @cache  # noqa: B019
    def copy_embeddings_query(self, destination: TableDestination) -> sql.Composed:
        return sql.SQL(
            "COPY {} ({}, chunk_seq, chunk, embedding) FROM STDIN WITH (FORMAT BINARY)"
        ).format(self.target_table_ident(destination), self.pk_fields_sql)  # type: ignore

    @cache  # noqa: B019
    def update_embedding_query(self, destination: ColumnDestination) -> sql.Composed:
        """Returns a SQL query to update the embedding column (for ColumnDestination)"""
        return sql.SQL("UPDATE {} SET {} = %s WHERE ({}) = ({})").format(
            self.source_table_ident,
            sql.Identifier(destination.embedding_column),
            self.pk_fields_sql,
            sql.SQL(", ").join([sql.Placeholder() for _ in self.pk_fields]),
        )

    @cached_property
    def insert_errors_query(self) -> sql.Composed:
        return sql.SQL(
            "INSERT INTO {} (id, message, details) VALUES (%s, %s, %s)"
        ).format(
            self.errors_table_ident,
        )

    def _pks_placeholders_tuples(self, items_count: int) -> sql.Composed:
        """Generates a comma separated list of tuples with placeholders for the
        primary key fields of the source table.

        If the primary key has 2 fields, and we want to generate the
        placeholders to match against 3 items:

        self._pks_placeholders_tuples(3)
        # => "(%s, %s), (%s, %s), (%s, %s)"

        This can be used for queries like:

        DELETE FROM table WHERE (pk1, pk2) IN ((%s, %s), (%s, %s), (%s, %s))

        We cannot use ANY = %s because Postgres doesn't allow it for anonymous
        composite values.
        """
        placeholder_tuple = sql.SQL(", ").join(
            sql.Placeholder() for _ in range(len(self.vectorizer.source_pk))
        )

        tuples = sql.SQL(",").join(
            sql.SQL("({})").format(placeholder_tuple) for _ in range(items_count)
        )
        return tuples

    @cached_property
    def is_vectorizer_disabled_query(self) -> sql.Composed:
        return sql.SQL("SELECT disabled FROM {} WHERE id = %s").format(
            self.vectorizer_table_ident,
        )

    @cached_property
    def reinsert_work_to_retry_query(self) -> sql.Composed:
        return sql.SQL("""
            INSERT INTO {queue_table}
                ({pk_fields}, loading_retries, loading_retry_after)
            VALUES
                ({pk_values}, (%(loading_retries)s+1),
                now() + INTERVAL '3 minutes'* (%(loading_retries)s + 1))
                        """).format(
            pk_fields=self.pk_fields_sql,
            queue_table=sql.Identifier(
                self.vectorizer.queue_schema, self.vectorizer.queue_table
            ),
            pk_values=sql.SQL(",").join(
                sql.SQL("(%(pk{})s)").format(sql.Literal(i))
                for i in range(len(self.pk_fields))
            ),
        )

    @cached_property
    def insert_queue_failed_query(self) -> sql.Composed:
        return sql.SQL("""
            INSERT INTO {queue_failed_table}
                ({pk_fields}, failure_step)
            VALUES
                ({pk_values}, %(failure_step)s)""").format(
            queue_failed_table=sql.Identifier(
                self.vectorizer.queue_schema,
                self.vectorizer.queue_failed_table,  # type: ignore
            ),
            pk_fields=self.pk_fields_sql,
            pk_values=sql.SQL(",").join(
                sql.SQL("(%(pk{})s)").format(sql.Literal(i))
                for i in range(len(self.pk_fields))
            ),
        )


class ProcessingStats:
    """
    Tracks processing statistics for chunk processing tasks.

    Records the total processing time, number of chunks processed,
    and wall time. It also logs statistics such as chunks processed per second.

    Attributes:
        total_processing_time (float): The total time spent processing chunks.
        total_chunks (int): The total number of chunks processed.
        wall_time (float): The total wall time from when processing started.
        wall_start (float): The time when processing started, used for
            calculating the elapsed time.
    """

    total_processing_time: float
    total_chunks: int
    wall_time: float
    wall_start: float

    def __new__(cls):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)

            logger.debug(
                "ProcessingStats initialized",
                thread_id=threading.get_native_id(),
                active_count=threading.active_count(),
                task=id(asyncio.current_task()),
            )
            cls._instance.total_processing_time = 0.0
            cls._instance.total_chunks = 0
            cls._instance.wall_start = time.perf_counter()
        return cls._instance

    def add_request_time(self, duration: float, chunk_count: int):
        """
        Adds the time and chunk count of a processing request to the accumulated totals.

        Args:
            duration (float): The time taken for the request.
            chunk_count (int): The number of chunks processed in the request.
        """
        self.total_processing_time += duration
        self.total_chunks += chunk_count

    async def print_stats(self):
        """
        Logs the processing statistics, including total time, chunks processed,
        and chunks processed per second.

        This method calculates two main rates:
        - Chunks processed per second overall.
        - Chunks processed per second per thread.

        The statistics are logged only to the DEBUG log level.
        """
        chunks_per_second_per_thread = (
            self.total_chunks / self.total_processing_time
            if self.total_processing_time > 0
            else 0
        )
        wall_time = time.perf_counter() - self.wall_start
        chunks_per_second = self.total_chunks / wall_time if wall_time > 0 else 0
        await logger.adebug(
            "Processing stats",
            wall_time=wall_time,
            total_processing_time=self.total_processing_time,
            total_chunks=self.total_chunks,
            chunks_per_second=chunks_per_second,
            chunks_per_second_per_thread=chunks_per_second_per_thread,
            task=id(asyncio.current_task()),
        )


T = TypeVar("T")


def flexible_take(iterable: list[T]) -> Callable[[int], list[T]]:
    """Creates a function that takes n elements from the iterable each time
    it's called."""
    # Convert to iterator if it's not already one
    iterator = iter(iterable)

    def take(n: int):
        # Convert to list so we get exactly what was requested
        return list(islice(iterator, n))

    return take


class UUIDEncoder(json.JSONEncoder):
    """A JSON encoder which can dump UUID."""

    @override
    def default(self, o: Any):
        if isinstance(o, UUID):
            return str(o)
        return json.JSONEncoder.default(self, o)


class Executor:
    """
    Responsible for processing items from the work queue and generating embeddings.

    The Executor fetches tasks from a database queue table, processes them using
    the vectorizer, and writes the resulting embeddings or errors back to the
    database.

    Attributes:
        db_url (str): The URL of the database to connect to.
        vectorizer (Vectorizer): The vectorizer configuration used for processing.
        queries (VectorizerQueryBuilder): A query builder instance used for
            generating SQL queries.
    """

    _queue_table_oid = None
    _should_continue_processing_hook: Callable[[int, int], bool]

    def __init__(
        self,
        db_url: str,
        vectorizer: Vectorizer,
        features: Features,
        worker_tracking: WorkerTracking,
        should_continue_processing_hook: None | Callable[[int, int], bool] = None,
    ):
        self.db_url = db_url
        self.vectorizer = vectorizer
        self.queries = VectorizerQueryBuilder(vectorizer)
        self._should_continue_processing_hook = should_continue_processing_hook or (
            lambda _loops, _res: True
        )
        self.features = features
        self.copy_types: None | Sequence[int] = None
        self.worker_tracking = worker_tracking

    async def run(self) -> int:
        """
        Embedding loop. Continuously fetches tasks from the work queue and
        processes them within the context of a transaction.

        Returns:
            int: The number of tasks processed from the work queue.
        """
        res = 0
        loops = 0

        async with await psycopg.AsyncConnection.connect(
            self.db_url,
            autocommit=True,
            application_name=f"pgai-worker[{self.vectorizer.id}]: {self.worker_tracking.get_short_worker_id()}",  # noqa: E501
        ) as conn:
            try:
                set_json_dumps(partial(json.dumps, cls=UUIDEncoder), context=conn)
                await register_vector_async(conn)
                await self.vectorizer.config.embedding.setup()
                while True:
                    if not await self._should_continue_processing(conn, loops, res):
                        return res
                    items_processed = await self._do_batch(conn)
                    if items_processed == 0:
                        return res
                    res += items_processed
                    loops += 1
                    await self.worker_tracking.save_vectorizer_success(
                        conn, self.vectorizer.id, items_processed
                    )
            except EmbeddingProviderError as e:
                async with conn.transaction():
                    await self._insert_vectorizer_error(
                        conn,
                        (
                            self.vectorizer.id,
                            e.msg,
                            Jsonb(
                                {
                                    "provider": self.vectorizer.config.embedding.implementation,
                                    "error_reason": str(e.__cause__),
                                }
                            ),
                        ),
                    )

                if e.__cause__ is not None:
                    raise e.__cause__  # noqa: B904
                raise e

            except Exception as e:
                async with conn.transaction():
                    await self._insert_vectorizer_error(
                        conn,
                        (
                            self.vectorizer.id,
                            VECTORIZER_FAILED,
                            Jsonb({"error_reason": str(e)}),
                        ),
                    )
                raise e

    async def _should_continue_processing(
        self, conn: AsyncConnection, loops: int, res: int
    ) -> bool:
        """Determine whether to continue processing based on vectorizer status
        and custom logic.

        Args:
            conn (AsyncConnection): Database connection for checking vectorizer status.
            loops (int): Number of processing loops completed.
            res (int): Result from the previous processing iteration.

        Returns:
            bool: True if processing should continue, False otherwise.

        Raises:
            Exception: If the vectorizer row is not found in the database.
        """
        if self.features.disable_vectorizers:
            async with conn.cursor() as cursor:
                await cursor.execute(
                    self.queries.is_vectorizer_disabled_query,
                    [self.vectorizer.id],
                )
                row = await cursor.fetchone()
                if not row:
                    raise Exception("vectorizer row not found")
                if row[0]:
                    return False

        return self._should_continue_processing_hook(loops, res)

    @tracer.wrap()
    async def _do_batch(self, conn: AsyncConnection) -> int:
        """
        Processes a batch of tasks. Fetches items from the queue, filters out
        deleted items, generates embeddings, and writes them to the database.

        Args:
            conn (AsyncConnection): The asynchronous database connection.

        Returns:
            int: The number of items processed in the batch.
        """
        processing_stats = ProcessingStats()
        start_time = time.perf_counter()
        async with conn.transaction():
            items = await self._fetch_work(conn)

            current_span = tracer.current_span()
            if current_span:
                current_span.set_tag("items_from_queue.pulled", len(items))
            await logger.adebug(f"Items pulled from queue: {len(items)}")

            # Filter out items that were deleted from the source table.
            # We use the first primary key column, since they can only
            # be null if the LEFT JOIN didn't find a match.
            items = [
                i for i in items if i[self.vectorizer.source_pk[0].attname] is not None
            ]

            if len(items) == 0:
                return 0

            num_chunks = await self._embed_and_write(conn, items)

            processing_stats.add_request_time(
                time.perf_counter() - start_time, num_chunks
            )
            await processing_stats.print_stats()

            return len(items)

    @cached_property
    def _batch_size(self) -> int:
        """Returns the batch size for processing.
        Documents take way longer to process than simple text rows,
        due to download and parsing overhead.
        So when the vectorizer is processing documents
        we use a smaller default batch size."""
        if self.vectorizer.config.processing.batch_size is not None:
            return max(1, min(self.vectorizer.config.processing.batch_size, 2048))
        else:
            if isinstance(self.vectorizer.config.loading, UriLoading):
                return 1
            else:
                return 50

    async def _fetch_work(self, conn: AsyncConnection) -> list[SourceRow]:
        """
        Fetches a batch of tasks from the work queue table. Safe for concurrent use.

        Follows the approach described in:
        https://www.timescale.com/blog/how-we-designed-a-resilient-vector-embedding-creation-system-for-postgresql-data/

        Args:
            conn (AsyncConnection): The database connection.

        Returns:
            list[SourceRow]: The rows from the source table that need to be embedded.
        """
        queue_table_oid = await self._get_queue_table_oid(conn)
        async with conn.cursor(row_factory=dict_row) as cursor:
            if self.features.loading_retries:
                await cursor.execute(
                    self.queries.fetch_work_query_with_retries,
                    (
                        self._batch_size,
                        queue_table_oid,
                    ),
                )
            else:
                await cursor.execute(
                    self.queries.fetch_work_query,
                    (
                        self._batch_size,
                        queue_table_oid,
                    ),
                )
            return await cursor.fetchall()

    async def _get_queue_table_oid(self, conn: AsyncConnection) -> int:
        """
        Retrieves the OID (Object Identifier) of the queue table.

        Args:
            conn (AsyncConnection): The database connection.

        Returns:
            int: The OID of the queue table.
        """
        if self._queue_table_oid is not None:
            return self._queue_table_oid

        async with conn.cursor(row_factory=dict_row) as cursor:
            await cursor.execute(
                self.queries.fetch_queue_table_oid_query,
            )
            row = await cursor.fetchone()
            if not row:
                raise Exception("work queue table doesn't exist")
            self._queue_table_oid = row["to_regclass"]
        return self._queue_table_oid

    @tracer.wrap()
    async def _embed_and_write(self, conn: AsyncConnection, items: list[SourceRow]):
        """
        Embeds the items and writes them to the database.

        - Deletes existing embeddings for the items.
        - Generates the documents to be embedded, chunks them, and formats the chunks.
        - Sends the documents to the embedding provider and writes embeddings
          to the database.
        - Logs any non-fatal errors encountered during embedding.

        Args:
            conn (AsyncConnection): The database connection.
            items (list[SourceRow]): The items to be embedded.

        Returns:
            int: The number of records written to the database.
        """

        await self._delete_embeddings(conn, items)
        count = 0
        async for records, loading_errors in self._generate_embeddings(items):
            if loading_errors:
                await self.handle_loading_retries(conn, loading_errors)
            await self._write_embeddings(conn, records)
            count += len(records)
        return count

    async def _delete_embeddings(self, conn: AsyncConnection, items: list[SourceRow]):
        """
        Deletes the embeddings for the given items from the target table.

        Args:
            conn (AsyncConnection): The database connection.
            items (list[SourceRow]): The items whose embeddings need to be deleted.
        """
        if self.vectorizer.config.destination.implementation == "column":
            return
        else:
            ids = [item[pk] for item in items for pk in self.queries.pk_attnames]
            async with conn.cursor() as cursor:
                await cursor.execute(
                    self.queries.delete_embeddings_query(
                        len(items), self.vectorizer.config.destination
                    ),
                    ids,
                )

    async def _load_copy_types(
        self, conn: AsyncConnection, destination: TableDestination
    ) -> None:
        """
        Loads the database types for the columns of the target table into
        self.copy_types.

        Args:
            conn (AsyncConnection): The database connection.
        """
        target_schema = destination.target_schema
        target_table = destination.target_table

        target_columns: list[str] = list(self.queries.pk_attnames) + [
            "chunk_seq",
            "chunk",
            "embedding",
        ]

        async with conn.cursor() as cursor:
            await cursor.execute(
                """
                select a.attname, a.atttypid
                from pg_catalog.pg_class k
                inner join pg_catalog.pg_namespace n
                    on (k.relnamespace operator(pg_catalog.=) n.oid)
                inner join pg_catalog.pg_attribute a
                    on (k.oid operator(pg_catalog.=) a.attrelid)
                where n.nspname operator(pg_catalog.=) %s
                and k.relname operator(pg_catalog.=) %s
                AND a.attname = ANY(%s)
                and a.attnum operator(pg_catalog.>) 0
            """,
                (
                    target_schema,
                    target_table,
                    target_columns,
                ),
            )
            column_name_to_type = {row[0]: row[1] for row in await cursor.fetchall()}
            self.copy_types = [column_name_to_type[col] for col in target_columns]
        assert self.copy_types is not None
        # len(source_pk) + chunk_seq + chunk + embedding
        assert len(self.copy_types) == len(self.vectorizer.source_pk) + 3

    @tracer.wrap()
    async def _copy_embeddings(
        self,
        conn: AsyncConnection,
        records: list[EmbeddingRecord],
        destination: TableDestination,
    ):
        """
        Inserts embeddings into the target table.

        For TableDestination, uses COPY FROM STDIN WITH (FORMAT BINARY).
        For ColumnDestination, uses UPDATE statements for each row.

        Args:
            conn (AsyncConnection): The database connection.
            records (list[EmbeddingRecord]): The embedding records to be copied.
        """
        if self.copy_types is None:
            await self._load_copy_types(conn, destination)
        async with (
            conn.cursor(binary=True) as cursor,
            cursor.copy(self.queries.copy_embeddings_query(destination)) as copy,  # type: ignore
        ):
            assert self.copy_types is not None  # ugh. make pyright happy
            copy.set_types(self.copy_types)
            for record in records:
                await copy.write_row(record)

        assert self.copy_types is not None
        # len(source_pk) + chunk_seq + chunk + embedding
        assert len(self.copy_types) == len(self.vectorizer.source_pk) + 3

    async def _update_source_table(
        self,
        destination: ColumnDestination,
        conn: AsyncConnection,
        records: list[EmbeddingRecord],
    ):
        async with conn.cursor() as cursor:
            for record in records:
                pk_values = record[: len(self.queries.pk_attnames)]
                embedding = record[-1]  # Last item is the embedding
                await cursor.execute(
                    self.queries.update_embedding_query(destination),  # type: ignore
                    [embedding] + pk_values,
                )

    async def _write_embeddings(
        self, conn: AsyncConnection, records: list[EmbeddingRecord]
    ):
        if self.vectorizer.config.destination.implementation == "table":
            await self._copy_embeddings(
                conn, records, self.vectorizer.config.destination
            )

        if self.vectorizer.config.destination.implementation == "column":
            await self._update_source_table(
                self.vectorizer.config.destination, conn, records
            )

    async def _insert_vectorizer_error(
        self,
        conn: AsyncConnection,
        record: VectorizerErrorRecord,
    ):
        """
        Inserts a single vectorizer error into the errors table.

        Args:
            conn (AsyncConnection): The database connection.
            record (VectorizerErrorRecord): The error record to be inserted.
        """
        async with conn.cursor() as cursor:
            await cursor.execute(self.queries.insert_errors_query, record)

    def _get_item_pk_values(self, item: SourceRow) -> list[Any]:
        return [item[pk] for pk in self.queries.pk_attnames]

    async def _generate_embeddings(
        self, items: list[SourceRow]
    ) -> AsyncGenerator[
        tuple[list[EmbeddingRecord], list[tuple[SourceRow, LoadingError]]], None
    ]:
        """
        Generates the embeddings for the given items.

        Args:
            items (list[SourceRow]): The items to generate embeddings for.

        Returns:
            AsyncGenerator[
                tuple[
                    list[EmbeddingRecord],
                    list[tuple[SourceRow, LoadingError]],
                ], None
            ]: A tuple of embedding records and error records.
        """
        # Note: deferred import to avoid import overhead
        import numpy as np

        records_without_embeddings: list[EmbeddingRecord] = []
        loading_errors: list[tuple[SourceRow, LoadingError]] = []
        documents: list[str] = []
        for item in items:
            pk_values = self._get_item_pk_values(item)
            try:
                payload = self.vectorizer.config.loading.load(item)
            except Exception as e:
                if self.features.loading_retries:
                    loading_errors.append((item, (LoadingError(e=e))))
                continue

            if isinstance(payload, str) and not payload:
                continue

            payload = self.vectorizer.config.parsing.parse(item, payload)
            chunks = self.vectorizer.config.chunking.into_chunks(item, payload)
            for chunk_id, chunk in enumerate(chunks, 0):
                formatted = self.vectorizer.config.formatting.format(chunk, item)
                records_without_embeddings.append(pk_values + [chunk_id, formatted])
                documents.append(formatted)

        if not documents:
            yield [], []

        if loading_errors:
            yield [], loading_errors

        try:
            rwe_take = flexible_take(records_without_embeddings)
            async for embeddings in self.vectorizer.config.embedding.embed(documents):
                records: list[EmbeddingRecord] = []
                for record, embedding in zip(
                    rwe_take(len(embeddings)), embeddings, strict=True
                ):
                    records.append(record + [np.array(embedding)])
                yield records, []
        except Exception as e:
            raise EmbeddingProviderError() from e

    async def handle_loading_retries(
        self,
        conn: AsyncConnection,
        loading_errors: list[tuple[SourceRow, LoadingError]],
    ):
        for item, e in loading_errors:
            is_retryable = await self._reinsert_loading_work_to_retry(
                conn, self.vectorizer.config.loading.retries, item
            )
            await self._insert_vectorizer_error(
                conn,
                (
                    self.vectorizer.id,
                    e.msg,
                    Jsonb(
                        {
                            "loader": self.vectorizer.config.loading.implementation,  # noqa
                            "error_reason": str(e.__cause__),
                            "is_retryable": is_retryable,
                        }
                    ),
                ),
            )

    async def _reinsert_loading_work_to_retry(
        self, conn: AsyncConnection, max_loading_retries: int, item: SourceRow
    ) -> bool:
        """
        Requeue the work items that failed to generate embeddings.
        """

        loading_retries = item.get("loading_retries", 0)
        if loading_retries >= max_loading_retries:
            queue_failed_params = {
                "failure_step": "loading",
                **{
                    f"pk{i}": value
                    for i, value in enumerate(self._get_item_pk_values(item))
                },
            }

            async with conn.cursor() as cursor:
                await cursor.execute(
                    self.queries.insert_queue_failed_query, queue_failed_params
                )
            return False

        reinsert_params = {
            "loading_retries": loading_retries,
            **{
                f"pk{i}": value
                for i, value in enumerate(self._get_item_pk_values(item))
            },
        }

        async with conn.cursor() as cursor:
            await cursor.execute(
                self.queries.reinsert_work_to_retry_query,
                reinsert_params,
            )
            return True


TIKTOKEN_CACHE_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "tiktoken_cache"
)



import asyncio
import os
from pathlib import Path

import psycopg
import pytest

import pgai
import pgai.semantic_catalog.search as search
from pgai.semantic_catalog import semantic_catalog
from pgai.semantic_catalog.vectorizer import (
    OllamaConfig,
    OpenAIConfig,
    SentenceTransformersConfig,
    vectorize_query,
)
from tests.semantic_catalog.utils import PostgresContainer

DATABASE = "sc_04"


@pytest.fixture(scope="module")
def event_loop():
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="module", autouse=True)
async def setup_database(container: PostgresContainer) -> None:
    container.drop_database(DATABASE, force=True)
    container.create_database(DATABASE)
    pgai.install(container.connection_string(database=DATABASE))
    script = Path(__file__).parent.joinpath("data", "vector_dump.sql").read_text()
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        # create a default semantic catalog
        sc = await semantic_catalog.create(con)
        # add a second embedding config
        await sc.add_embedding(
            con,
            OpenAIConfig.create(
                model="text-embedding-3-small",
                dimensions=1536,
            ),
        )
        # add a third embedding config
        await sc.add_embedding(
            con,
            OllamaConfig.create(
                model="nomic-embed-text",
                dimensions=768,
            ),
        )
        # load descriptions+vectors into the semantic catalog
        async with con.cursor() as cur:
            await cur.execute(script)  # pyright: ignore [reportArgumentType]


async def test_list_objects(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        obj_descs = await sc.list_objects(con)
        assert len(obj_descs) > 0
        obj_desc = obj_descs[0]
        assert obj_desc.objnames == ["postgres_air", "airport"]
        assert obj_desc.description is not None


async def test_list_objects_objtype(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        obj_descs = await sc.list_objects(con, objtype="table column")
        assert len(obj_descs) > 0
        obj_desc = obj_descs[0]
        assert obj_desc.objnames == ["postgres_air", "airport", "airport_code"]
        assert obj_desc.description is not None


async def test_search_obj_sentence_transformers(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb1")
        assert isinstance(emb_cfg, SentenceTransformersConfig)
        vec = await vectorize_query(
            emb_cfg, "What column will tell me if an airport is international?"
        )
        obj_descs = await search.search_objects(con, sc.id, "emb1", emb_cfg, vec)
        assert len(obj_descs) > 0
        obj_desc = obj_descs[0]
        assert obj_desc.objnames == ["postgres_air", "airport", "intnl"]
        assert obj_desc.description is not None


async def test_search_obj_openai(container: PostgresContainer) -> None:
    if "OPENAI_API_KEY" not in os.environ:
        pytest.skip("OPENAI_API_KEY is not set")
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb2")
        assert isinstance(emb_cfg, OpenAIConfig)
        vec = await vectorize_query(
            emb_cfg, "What column will tell me if an airport is international?"
        )
        obj_descs = await search.search_objects(con, sc.id, "emb2", emb_cfg, vec)
        assert len(obj_descs) > 0
        obj_desc = obj_descs[0]
        assert obj_desc.objnames == ["postgres_air", "airport", "intnl"]
        assert obj_desc.description is not None


async def test_search_obj_ollama(container: PostgresContainer) -> None:
    if "OLLAMA_HOST" not in os.environ:
        pytest.skip("OLLAMA_HOST is not set")
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb3")
        assert isinstance(emb_cfg, OllamaConfig)
        vec = await vectorize_query(
            emb_cfg, "What column will tell me if an airport is international?"
        )
        obj_descs = await search.search_objects(con, sc.id, "emb3", emb_cfg, vec)
        assert len(obj_descs) > 0
        obj_desc = obj_descs[0]
        assert obj_desc.objnames == ["postgres_air", "airport", "intnl"]
        assert obj_desc.description is not None


async def test_list_sql(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        sql_examples = await sc.list_sql_examples(con)
        assert len(sql_examples) > 0
        sql_example = sql_examples[0]
        assert (
            sql_example.description == "Delayed flights are indicated by their status"
        )


async def test_search_sql_sentence_transformers(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb1")
        assert isinstance(emb_cfg, SentenceTransformersConfig)
        vec = await vectorize_query(emb_cfg, "Is my flight delayed?")
        sql_examples = await search.search_sql_examples(
            con, sc.id, "emb1", emb_cfg, vec
        )
        assert len(sql_examples) > 0
        sql_example = sql_examples[0]
        assert (
            sql_example.description == "Delayed flights are indicated by their status"
        )


async def test_search_sql_openai(container: PostgresContainer) -> None:
    if "OPENAI_API_KEY" not in os.environ:
        pytest.skip("OPENAI_API_KEY is not set")
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb2")
        assert isinstance(emb_cfg, OpenAIConfig)
        vec = await vectorize_query(emb_cfg, "Is my flight delayed?")
        sql_examples = await search.search_sql_examples(
            con, sc.id, "emb2", emb_cfg, vec
        )
        assert len(sql_examples) > 0
        sql_example = sql_examples[0]
        assert (
            sql_example.description == "Delayed flights are indicated by their status"
        )


async def test_search_sql_ollama(container: PostgresContainer) -> None:
    if "OLLAMA_HOST" not in os.environ:
        pytest.skip("OLLAMA_HOST is not set")
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb3")
        assert isinstance(emb_cfg, OllamaConfig)
        vec = await vectorize_query(emb_cfg, "Is my flight delayed?")
        sql_examples = await search.search_sql_examples(
            con, sc.id, "emb3", emb_cfg, vec
        )
        assert len(sql_examples) > 0
        sql_example = sql_examples[0]
        assert (
            sql_example.description == "Delayed flights are indicated by their status"
        )


async def test_list_facts(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        facts = await sc.list_facts(con)
        assert len(facts) > 0
        fact = facts[0]
        assert (
            fact.description
            == "The postgres_air.airport.iso_region values are in uppercase."
        )


async def test_search_fact_sentence_transformers(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb1")
        assert isinstance(emb_cfg, SentenceTransformersConfig)
        vec = await vectorize_query(emb_cfg, "What airports are in the Texas region?")
        facts = await search.search_facts(con, sc.id, "emb1", emb_cfg, vec)
        assert len(facts) > 0
        fact = facts[0]
        assert (
            fact.description
            == "The postgres_air.airport.iso_region values are in uppercase."
        )


async def test_search_fact_openai(container: PostgresContainer) -> None:
    if "OPENAI_API_KEY" not in os.environ:
        pytest.skip("OPENAI_API_KEY is not set")
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb2")
        assert isinstance(emb_cfg, OpenAIConfig)
        vec = await vectorize_query(emb_cfg, "What airports are in the Texas region?")
        facts = await search.search_facts(con, sc.id, "emb2", emb_cfg, vec)
        assert len(facts) > 0
        fact = facts[0]
        assert (
            fact.description
            == "The postgres_air.airport.iso_region values are in uppercase."
        )


async def test_search_fact_ollama(container: PostgresContainer) -> None:
    if "OLLAMA_HOST" not in os.environ:
        pytest.skip("OLLAMA_HOST is not set")
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        emb_cfg = await sc.get_embedding(con, "emb3")
        assert isinstance(emb_cfg, OllamaConfig)
        vec = await vectorize_query(emb_cfg, "What airports are in the Texas region?")
        facts = await search.search_facts(con, sc.id, "emb3", emb_cfg, vec)
        assert len(facts) > 0
        fact = facts[0]
        assert (
            fact.description
            == "The postgres_air.airport.iso_region values are in uppercase."
        )
